from typing import Optional, Union, Match, Dict, List
import re

MASK = "***"
DEFAULT_MASKING_FIELDS = {"token": 0, "access_token": 0, "refresh_token": 0, "epkId": 0, "profileId": 0}
CARD_MASKING_FIELDS = ["message", "debug_info", "normalizedMessage", "incoming_text", "annotations", "inner_entities",
                       "preprocess_result", "original_message", "original_tokenized_elements"]

card_regular = re.compile(r"(?:(\d{18})|(\d{16})|(?:\d{4} ){3}(\d{4})(\s?\d{2})?)")


class Counter(object):
    # Класс счетчик для маскировки структуры, items - кол-во простых элементов, collections - коллекций
    # max_depth - максимальная глубина
    def __init__(self):
        self.items = 0
        self.collections = 0
        self.max_depth = 0


def luhn_checksum(card_number: str) -> bool:
    digits = [int(d) for d in card_number]
    odd_digits = digits[-1::-2]
    even_digits = digits[-2::-2]
    checksum = sum(odd_digits) + sum(map(lambda x: (x * 2) % 10 + (x * 2) // 10, even_digits))
    return checksum % 10 == 0


def card_sub_func(x: Match[str]) -> str:
    d_regular = re.compile(r"\d")

    g0 = x.group(0)
    is_last_not_digit = int(g0 and not g0[-1].isdigit())
    last_char = g0[-1]

    mask = d_regular.sub("*", x.group(0))[:-(4 + is_last_not_digit)]
    digs = (x.group(0) or '').replace(' ', '')[-4:]
    return mask + digs + (last_char * is_last_not_digit)


def masking(data: Union[Dict, List], masking_fields: Optional[Union[Dict, List]] = None,
            depth_level: int = 2, mask_available_depth: int = -1) -> Union[Dict, List]:
    """
    :param data: коллекция для маскирования приватных данных
    :param masking_fields: поля для обязательной маскировки независимо от уровня
    :param depth_level: глубина сохранения структуры маскируемого поля
    :param mask_available_depth: глубина глубокой маскировки полей без сохранения структуры (см ниже)
    """

    if isinstance(masking_fields, list):
        masking_fields = {key: depth_level for key in masking_fields}

    if masking_fields is None:
        masking_fields = DEFAULT_MASKING_FIELDS

    return _masking(data, masking_fields, depth_level, mask_available_depth, masking_on=False, card_masking_on=False)


def _masking(data: Union[Dict, List], masking_fields: Union[Dict, List],
             depth_level: int = 2, mask_available_depth: int = -1, masking_on: bool = False,
             card_masking_on: bool = False) -> Union[Dict, List]:

    # тут в зависимости от листа или словаря создаем итератор
    if isinstance(data, dict):
        key_gen = data.items()
        masked_data = dict()
    else:
        key_gen = enumerate(data)
        masked_data = [None for _ in range(len(data))]

    for key, _ in key_gen:
        value_is_collection = isinstance(data[key], (dict, list))
        if isinstance(data[key], (set, tuple)):
            data[key] = list(data[key])
            value_is_collection = True
        if masking_on or key in masking_fields:
            if value_is_collection:
                # если глубина не превышена, идем внутрь с включенным флагом и уменьшаем глубину
                if masking_on and depth_level > 0:
                    masked_data[key] = _masking(data[key], masking_fields, depth_level - 1,
                                                mask_available_depth, masking_on=True)
                elif key in masking_fields and masking_fields[key] > 0:
                    masked_data[key] = _masking(data[key], masking_fields, masking_fields[key] - 1,
                                                mask_available_depth, masking_on=True)
                else:
                    counter = structure_mask(data[key], depth=1, available_depth=mask_available_depth)
                    masked_data[key] = f'*items-{counter.items}*collections-{counter.collections}*maxdepth-{counter.max_depth}*'
            elif data[key] is not None:  # в случае простого элемента. маскируем как ***
                masked_data[key] = '***'
            else:
                masked_data[key] = data[key]
        elif key in CARD_MASKING_FIELDS or card_masking_on:  # проверка на реквизиты карты
            if value_is_collection:
                masked_data[key] = _masking(data[key], masking_fields, depth_level,
                                            mask_available_depth, masking_on, card_masking_on=True)
            elif isinstance(data[key], str):
                masked_data[key] = card_regular.sub(card_sub_func, data[key])
            elif isinstance(data[key], int):
                str_value = str(data[key])
                masked_value = card_regular.sub(card_sub_func, str_value)
                if masked_value != str_value:
                    masked_data[key] = masked_value
                else:
                    masked_data[key] = data[key]
            else:
                masked_data[key] = data[key]
        elif value_is_collection:
            # если маскировка не нужна уходим глубже без включенного флага
            masked_data[key] = _masking(data[key], masking_fields, depth_level, mask_available_depth,
                                        masking_on=False, card_masking_on=card_masking_on)
        else:
            masked_data[key] = data[key]
    return masked_data


def structure_mask(data: Union[Dict, List], depth: int, available_depth: int = -1,
                   counter: Optional[Counter] = None):
    """
    Функция максировки для сложной структуры
    :param data: структура маскируемая без сохранения структуры
    :param depth: текущая глубина вложенности
    :param available_depth: максимальная глубина прохода, при -1 глубина не ограничена
    :param counter: объект счетчик
    :return: counter с подсчитанными элементами
    """
    # в зависимости от листа или словаря создаем итератор
    if isinstance(data, dict):
        key_gen = data.items()
    else:
        key_gen = enumerate(data)

    if counter is None:
        counter = Counter()

    for key, _ in key_gen:
        if isinstance(data[key],(dict, list)):
            counter.collections += 1
            # если встречаем коллекцию и глубина не превышена идем внутрь
            if depth < available_depth or available_depth == -1:
                structure_mask(data[key], depth + 1, available_depth, counter)
        else:
            # если элемент простой крутим счетчик простых элементов
            counter.items += 1
            if depth > counter.max_depth:
                counter.max_depth = depth

    return counter



# coding: utf-8
from typing import List, Dict, Any

from core.basic_models.actions.command import Command
from core.logging.logger_utils import log
import scenarios.logging.logger_constants as log_const
from scenarios.user.user_model import User
from smart_kit.handlers.handler_base import HandlerBase
from smart_kit.message.app_info import AppInfo
from core.names.field import APP_INFO
from smart_kit.names.message_names import RUN_APP, MESSAGE_TO_SKILL, SERVER_ACTION
from core.monitoring.monitoring import monitoring


class HandlerTimeout(HandlerBase):
    def run(self, payload: Dict[str, Any], user: User) -> List[Command]:
        commands = super().run(payload, user)
        callback_id = user.message.callback_id
        if user.behaviors.has_callback(callback_id):
            params = {log_const.KEY_NAME: "handling_timeout"}
            log("TimeoutHandler started", user, params)
            action_params = user.behaviors.get_callback_action_params(callback_id)
            if action_params:
                app_info = None
                for original_message_name in [MESSAGE_TO_SKILL, SERVER_ACTION, RUN_APP]:
                    if original_message_name in action_params:
                        app_info = AppInfo(action_params[original_message_name].get(APP_INFO, {}))
                        break

                monitoring.counter_incoming(self.app_name, user.message.message_name, self.__class__.__name__,
                                            user, app_info=app_info)

            callback_id = user.message.callback_id
            commands.extend(user.behaviors.timeout(callback_id))
        return commands



from unittest import TestCase

from scenarios.scenario_models.field.field_filler_description import RegexpFieldFiller
from smart_kit.utils.picklable_mock import PicklableMock


class TestRegexpFiller(TestCase):
    def setUp(self):
        self.items = {"exp": "1-[0-9A-Z]{7}"}
        self.user = PicklableMock()
        self.user.message = PicklableMock()
        self.user.message.masked_value = ""

    def test_no_exp_init(self):
        self.assertRaises(KeyError, RegexpFieldFiller, {})

    def test_no_exp(self):
        field_value = "1-RSAR09A"
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.original_text = field_value

        filler = RegexpFieldFiller(self.items)
        filler.regexp = None
        self.assertIsNone(filler.extract(text_preprocessing_result, self.user))

    def test_extract(self):
        field_value = "1-RSAR09A"
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.original_text = field_value

        filler = RegexpFieldFiller(self.items)
        result = filler.extract(text_preprocessing_result, self.user)

        self.assertEqual(field_value, result)

    def test_extract_no_match(self):
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.original_text = "text"

        filler = RegexpFieldFiller(self.items)
        result = filler.extract(text_preprocessing_result, self.user)

        self.assertIsNone(result)

    def test_extract_mult_match_default_delimiter(self):
        field_value = "1-RSAR09A пустой тест 1-RSAR02A"
        res = ",".join(['1-RSAR09A', '1-RSAR02A'])
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.original_text = field_value

        filler = RegexpFieldFiller(self.items)
        result = filler.extract(text_preprocessing_result, self.user)

        self.assertEqual(res, result)

    def test_extract_mult_match_custom_delimiter(self):
        field_value = "1-RSAR09A пустой тест 1-RSAR02B"
        self.items["delimiter"] = ";"
        res = self.items["delimiter"].join(['1-RSAR09A', '1-RSAR02B'])
        text_preprocessing_result = PicklableMock()
        text_preprocessing_result.original_text = field_value

        filler = RegexpFieldFiller(self.items)
        result = filler.extract(text_preprocessing_result, self.user)

        self.assertEqual(res, result)



import unittest
from unittest.mock import Mock, patch

from smart_kit.action.http import HTTPRequestAction


class HttpRequestActionTest(unittest.TestCase):
    TIMEOUT = 4

    def setUp(self):
        self.user = Mock(
            parametrizer=Mock(collect=lambda *args, **kwargs: {}),
            descriptions={
                "behaviors": {
                    "my_behavior": Mock(timeout=Mock(return_value=self.TIMEOUT))
                }
            }
        )

    @staticmethod
    def set_request_mock_attribute(request_mock, return_value=None):
        return_value = return_value or {}
        request_mock.return_value = Mock(
            __enter__=Mock(return_value=Mock(
                json=Mock(return_value=return_value),
                cookies={},
                headers={},
            ), ),
            __exit__=Mock()
        )

    @patch('requests.request')
    def test_simple_request(self, request_mock: Mock):
        self.set_request_mock_attribute(request_mock, return_value={'data': 'value'})
        items = {
            "params": {
                "method": "POST",
                "url": "https://my.url.com",
            },
            "store": "user_variable",
            "behavior": "my_behavior",
        }
        HTTPRequestAction(items).run(self.user, None, {})
        request_mock.assert_called_with(url="https://my.url.com", method='POST', timeout=self.TIMEOUT)
        self.assertTrue(self.user.descriptions["behaviors"]["my_behavior"].success_action.run.called)
        self.assertTrue(self.user.variables.set.called)
        self.user.variables.set.assert_called_with("user_variable", {'data': 'value'})

    @patch('requests.request')
    def test_render_params(self, request_mock: Mock):
        self.set_request_mock_attribute(request_mock)
        items = {
            "params": {
                "method": "POST",
                "url": "https://{{url}}",
                "timeout": 3,
                "json": {
                    "param": "{{value}}"
                }
            },
            "store": "user_variable",
            "behavior": "my_behavior",
        }
        params = {
            "url": "my.url.com",
            "value": "my_value"
        }
        HTTPRequestAction(items).run(self.user, None, params)
        request_mock.assert_called_with(url="https://my.url.com", method='POST', timeout=3, json={"param": "my_value"})

    @patch('requests.request')
    def test_headers_fix(self, request_mock):
        self.set_request_mock_attribute(request_mock)
        items = {
            "params": {
                "headers": {
                    "header_1": 32,
                    "header_2": 32.03,
                    "header_3": b"d32",
                    "header_4": None,
                    "header_5": {"data": "value"},
                },
            },
            "store": "user_variable",
            "behavior": "my_behavior",
        }
        HTTPRequestAction(items).run(self.user, None, {})
        request_mock.assert_called_with(headers={
            "header_1": "32",
            "header_2": "32.03",
            "header_3": b"d32"
        }, method=HTTPRequestAction.DEFAULT_METHOD, timeout=self.TIMEOUT)

    @patch('requests.request')
    def test_beheivor_is_none(self, request_mock):
        self.set_request_mock_attribute(request_mock)
        items = {
            "params": {
            },
            "store": "user_variable",
        }
        HTTPRequestAction(items).run(self.user, None, {})
        request_mock.assert_called_with(method=HTTPRequestAction.DEFAULT_METHOD,
                                        timeout=HTTPRequestAction.DEFAULT_TIMEOUT)



import unittest
from unittest.mock import Mock

from core.model.lazy_items import LazyItems
from smart_kit.utils.picklable_mock import PicklableMock


class MockFactory:
    def __init__(self, id, items):
        self.id = id
        self.items = items


class MockDescriptions:
    def __init__(self, factory, items):
        self.items = items

    def __contains__(self, key):
        return key in self.items


class LazyItemsTest(unittest.TestCase):

    def test_clear_removed_items(self):
        raw_data = {}
        user = PicklableMock()
        description = PicklableMock()
        factory = Mock(raw_data, description, user)
        items = {"1": "test1", "2": "test2", "3": "test3"}
        raw_description = {"1": "test1", "2": "test2"}
        descriptions = MockDescriptions(factory, raw_description)
        items = LazyItems(items, descriptions, user, factory)
        self.assertEqual(items._raw_items, raw_description)



import unittest
from unittest.mock import patch, MagicMock

from core.basic_models.actions.smartpay import SmartPayCreateAction, SmartPayPerformAction, SmartPayGetStatusAction, \
    SmartPayConfirmAction, SmartPayDeleteAction, SmartPayRefundAction
from smart_kit.utils.picklable_mock import PicklableMock
from tests.smart_kit_tests.action.test_http import HttpRequestActionTest


class SmartPayActionTest(unittest.TestCase):
    def setUp(self):
        self.user = PicklableMock(
            parametrizer=PicklableMock(collect=lambda *args, **kwargs: {}),
            descriptions={
                "behaviors": {
                    "my_behavior": PicklableMock(timeout=PicklableMock(return_value=3))
                }
            }
        )
        self.settings = {
            "template_settings": {
                "smartpay_url": "0.0.0.0"
            }
        }

    @patch("smart_kit.configs.settings.Settings")
    @patch('requests.request')
    def test_create(self, request_mock: PicklableMock, settings_mock: MagicMock):
        items = {
            "behavior": "my_behavior",
            "smartpay_params": {
                "ptype": 1,
                "invoice": {
                    "purchaser": {
                        "email": "test@test.ru",
                        "phone": "71111111111",
                        "contact": "phone"
                    },
                    "delivery_info": {
                        "address": {
                            "country": "RU",
                            "city": "Москва",
                            "address": "Кутузовский проспект, 32"
                        },
                        "delivery_type": "courier",
                        "description": "Спросить '\''Где тут Сбердевайсы?'\''"
                    },
                    "invoice_params": [
                        {
                            "key": "packageName",
                            "value": "SbERdeVICEs"
                        },
                        {
                            "key": "packageKey",
                            "value": "ru.sberdevices.test"
                        }
                    ],
                    "order": {
                        "order_id": "a952b7ee-c928-4586-bd05-d932c21f749",
                        "order_number": "1952",
                        "order_date": "2021-04-07T08:24:37.729Z",
                        "service_id": "13",
                        "amount": 79900,
                        "currency": "RUB",
                        "purpose": "Покупка продуктов",
                        "description": "Заказ № 22-1952. Покупка продуктов",
                        "language": "ru-RU",
                        "tax_system": 0,
                        "order_bundle": [
                            {
                                "position_id": 1,
                                "name": "Транзистор",
                                "item_params": [
                                    {
                                        "key": "_itemAttributes_supplier_info.name",
                                        "value": "ООО Ромашка"
                                    },
                                    {
                                        "key": "_itemAttributes_supplier_info.inn",
                                        "value": "5009053292"
                                    },
                                    {
                                        "key": "_itemAttributes_nomenclature",
                                        "value": "13622200005881"
                                    },
                                    {
                                        "key": "_itemAttributes_agent_info.type",
                                        "value": "7"
                                    }

                                ],
                                "quantity": {
                                    "value": 1,
                                    "measure": "шт."
                                },
                                "item_amount": 79801,
                                "currency": "RUB",
                                "item_code": "21ff407a-fa98-11e8-80c5-0cc47a817926",
                                "item_price": 79801,
                                "discount_type": "amount",
                                "discount_value": 17687,
                                "tax_type": 6
                            },
                            {
                                "position_id": 2,
                                "name": "Тиристор",
                                "item_params": [
                                    {
                                        "key": "code",
                                        "value": "123значение321"
                                    },
                                    {
                                        "key": "pack",
                                        "value": "100"
                                    },
                                    {
                                        "key": "_itemAttributes_supplier_info.name",
                                        "value": "ООО Платежи"
                                    },
                                    {
                                        "key": "_itemAttributes_supplier_info.inn",
                                        "value": "7730253720"
                                    }
                                ],
                                "quantity": {
                                    "value": 1,
                                    "measure": "шт."
                                },
                                "item_amount": 99,
                                "currency": "RUB",
                                "item_code": "21ff407c-fa98-11e8-80c5-0cc47a817926",
                                "item_price": 99,
                                "discount_type": "amount",
                                "discount_value": 21,
                                "tax_type": 6,
                            }
                        ]
                    }
                }
            }
        }

        settings_mock.return_value = self.settings
        HttpRequestActionTest.set_request_mock_attribute(request_mock, return_value={'data': 'value'})
        action = SmartPayCreateAction(items)
        action.run(self.user, None, {})
        request_mock.assert_called_with(url="0.0.0.0/invoices", method="POST",
                                        timeout=3, json=items.get("smartpay_params"))
        self.assertTrue(self.user.descriptions["behaviors"]["my_behavior"].success_action.run.called)
        self.assertTrue(self.user.variables.set.called)
        self.user.variables.set.assert_called_with("smartpay_create_answer", {'data': 'value'})

    @patch("smart_kit.configs.settings.Settings")
    @patch('requests.request')
    def test_perform(self, request_mock: PicklableMock, settings_mock: MagicMock):
        items = {
            "behavior": "my_behavior",
            "invoice_id": "0",
            "smartpay_params": {
                "user_id": {
                    "partner_client_id": "2223"
                },
                "operations": [
                    {
                        "operation": "payment",
                        "code": "invoice",
                        "value": "167900"
                    }
                ],
                "device_info": {
                    "device_platform_type": "iOS",
                    "device_platform_version": "13.6.1",
                    "device_model": "iPhone 7",
                    "device_manufacturer": "Apple",
                    "device_id": "83c3f257-46d8-41fe-951b-f79d04e288c2",
                    "surface": "SBOL",
                    "surface_version": "11.5.0"
                },
                "return_url": "https://ok",
                "fail_url": "https://fail"
            }
        }

        settings_mock.return_value = self.settings
        HttpRequestActionTest.set_request_mock_attribute(request_mock, return_value={'data': 'value'})
        action = SmartPayPerformAction(items)
        action.run(self.user, None, {})
        request_mock.assert_called_with(url="0.0.0.0/invoices/0", method="POST",
                                        timeout=3, json=items.get("smartpay_params"))
        self.assertTrue(self.user.descriptions["behaviors"]["my_behavior"].success_action.run.called)
        self.assertTrue(self.user.variables.set.called)
        self.user.variables.set.assert_called_with("smartpay_perform_answer", {'data': 'value'})

    @patch("smart_kit.configs.settings.Settings")
    @patch('requests.request')
    def test_get_status(self, request_mock: PicklableMock, settings_mock: MagicMock):
        items = {
            "behavior": "my_behavior",
            "invoice_id": "0",
            "inv_status": "executed",
            "wait": 50
        }

        settings_mock.return_value = self.settings
        HttpRequestActionTest.set_request_mock_attribute(request_mock, return_value={'data': 'value'})
        action = SmartPayGetStatusAction(items)
        action.run(self.user, None, {})
        request_mock.assert_called_with(url="0.0.0.0/invoices/0", method="GET",
                                        timeout=3, params={"inv_status": "executed", "wait": 50})
        self.assertTrue(self.user.descriptions["behaviors"]["my_behavior"].success_action.run.called)
        self.assertTrue(self.user.variables.set.called)
        self.user.variables.set.assert_called_with("smartpay_get_status_answer", {'data': 'value'})

    @patch("smart_kit.configs.settings.Settings")
    @patch('requests.request')
    def test_partial_confirm(self, request_mock: PicklableMock, settings_mock: MagicMock):
        items = {
            "behavior": "my_behavior",
            "invoice_id": "0",
            "smartpay_params": {
                "invoice": {
                    "order": {
                        "amount": 79801,
                        "currency": "RUB",
                        "order_bundle": [
                            {
                                "position_id": 1,
                                "name": "Транзистор",
                                "quantity": {
                                    "value": 1,
                                    "measure": "шт."
                                },
                                "item_amount": 79801,
                                "currency": "RUB",
                                "item_code": "21ff407a-fa98-11e8-80c5-0cc47a817926",
                                "item_price": 79801
                            }
                        ]
                    }
                }
            }
        }

        settings_mock.return_value = self.settings
        HttpRequestActionTest.set_request_mock_attribute(request_mock, return_value={'data': 'value'})
        action = SmartPayConfirmAction(items)
        action.run(self.user, None, {})
        request_mock.assert_called_with(url="0.0.0.0/invoices/0", method="PUT",
                                        timeout=3, json=items["smartpay_params"])
        self.assertTrue(self.user.descriptions["behaviors"]["my_behavior"].success_action.run.called)
        self.assertTrue(self.user.variables.set.called)
        self.user.variables.set.assert_called_with("smartpay_confirm_answer", {'data': 'value'})

    @patch("smart_kit.configs.settings.Settings")
    @patch('requests.request')
    def test_full_confirm(self, request_mock: PicklableMock, settings_mock: MagicMock):
        items = {
            "behavior": "my_behavior",
            "invoice_id": "0",
        }

        settings_mock.return_value = self.settings
        HttpRequestActionTest.set_request_mock_attribute(request_mock, return_value={'data': 'value'})
        action = SmartPayConfirmAction(items)
        action.run(self.user, None, {})
        request_mock.assert_called_with(url="0.0.0.0/invoices/0", method="PUT", timeout=3)
        self.assertTrue(self.user.descriptions["behaviors"]["my_behavior"].success_action.run.called)
        self.assertTrue(self.user.variables.set.called)
        self.user.variables.set.assert_called_with("smartpay_confirm_answer", {'data': 'value'})

    @patch("smart_kit.configs.settings.Settings")
    @patch('requests.request')
    def test_delete(self, request_mock: PicklableMock, settings_mock: MagicMock):
        items = {
            "behavior": "my_behavior",
            "invoice_id": "0",
        }

        settings_mock.return_value = self.settings
        HttpRequestActionTest.set_request_mock_attribute(request_mock, return_value={'data': 'value'})
        action = SmartPayDeleteAction(items)
        action.run(self.user, None, {})
        request_mock.assert_called_with(url="0.0.0.0/invoices/0", method="DELETE", timeout=3)
        self.assertTrue(self.user.descriptions["behaviors"]["my_behavior"].success_action.run.called)
        self.assertTrue(self.user.variables.set.called)
        self.user.variables.set.assert_called_with("smartpay_delete_answer", {'data': 'value'})

    @patch("smart_kit.configs.settings.Settings")
    @patch('requests.request')
    def test_partial_refund(self, request_mock: PicklableMock, settings_mock: MagicMock):
        items = {
            "behavior": "my_behavior",
            "invoice_id": "0",
            "smartpay_params": {
                "invoice": {
                    "order": {
                        "current_amount": 79900,
                        "refund_amount": 79801,
                        "currency": "RUB",
                        "order_bundle": [
                            {
                                "position_id": 1,
                                "name": "Транзистор",
                                "quantity": {
                                    "value": 1
                                },
                                "item_amount": 79801,
                                "item_code": "21ff407a-fa98-11e8-80c5-0cc47a817926"
                            }
                        ]
                    }
                }
            }
        }

        settings_mock.return_value = self.settings
        HttpRequestActionTest.set_request_mock_attribute(request_mock, return_value={'data': 'value'})
        action = SmartPayRefundAction(items)
        action.run(self.user, None, {})
        request_mock.assert_called_with(url="0.0.0.0/invoices/0", method="PATCH",
                                        timeout=3, json=items["smartpay_params"])
        self.assertTrue(self.user.descriptions["behaviors"]["my_behavior"].success_action.run.called)
        self.assertTrue(self.user.variables.set.called)
        self.user.variables.set.assert_called_with("smartpay_refund_answer", {'data': 'value'})

    @patch("smart_kit.configs.settings.Settings")
    @patch('requests.request')
    def test_full_refund(self, request_mock: PicklableMock, settings_mock: MagicMock):
        items = {
            "behavior": "my_behavior",
            "invoice_id": "0",
        }

        settings_mock.return_value = self.settings
        HttpRequestActionTest.set_request_mock_attribute(request_mock, return_value={'data': 'value'})
        action = SmartPayRefundAction(items)
        action.run(self.user, None, {})
        request_mock.assert_called_with(url="0.0.0.0/invoices/0", method="PATCH", timeout=3)
        self.assertTrue(self.user.descriptions["behaviors"]["my_behavior"].success_action.run.called)
        self.assertTrue(self.user.variables.set.called)
        self.user.variables.set.assert_called_with("smartpay_refund_answer", {'data': 'value'})



from core.db_adapter.db_adapter import DBAdapterException


class NotSupportedOperation(DBAdapterException):
    def __init__(self, msg=None):
        msg = msg or "Operation is not supported"
        super().__init__(msg)



"""
Модуль для определения временного периода из текста на русском языке
!!! Модуль работает только со словами в нижнем регистре !!!

usage: begin_date, end_date = period_determiner(words_to_process)
"""

import re
from typing import List, Tuple, Optional
from datetime import datetime, timedelta


# глобальный формат даты
date_format: str = '%d.%m.%Y'
# шаблоны регулярок для определения дат
re_shortest_date_pattern = '^([0-9]{1,2})\\.([0-9]{1,2})$'
re_short_date_pattern = '^([0-9]{1,2})\\.([0-9]{1,2})\\.([0-9]{2})$'
re_long_date_pattern = '^([0-9]{1,2})\\.([0-9]{1,2})\\.([0-9]{4})$'
# константы для ошибки и не распознанной даты
ERROR_VALUE = 'error'
UNRECOGNIZED_DATE_VALUE = ''


class IncorrectDateException(Exception):
    pass


class StateMachineForDateDetermining:
    """
    Класс конечный-автомат (КА) для определения даты из строки на русском языке.
    Работает следующим образом:
    на вход в КА передаем последовательно слова, каждое новое слово влияет на состояние КА
    и КА понимает какой период времени пытаемся сообщить,
    а также КА может понять, что дата определена некорректно
    """

    # Истина, если используется период до текущего дня
    # иначе конкретная дата (день, месяц, квартал, год)
    _is_period: bool
    _is_determined: bool
    _is_error: bool
    _date_period: List[Optional[datetime]]
    _day: int
    _month: int
    _year: int

    # список следующих ожидаемых слов
    # если это слово не получено, то ошибка
    # например: если пришло "вчерашний", то ждем слово "день"
    # если любое следующее слово ошибка, как в случае текста "сегодня",
    # то _next_expected_words = []
    # если не рассматриваем следующие слова, то _next_expected_words = None
    _next_expected_words: Optional[List[str]]

    # дескриптор относительного периода:
    #  1 - указание на текущий период - нынешний, текущий, сегодняшний, этот
    #  0 - отсутствует
    # -1 - указание на прошлый период - вчерашний, вчера, прошлый, предыдущий
    _relative_descriptor: int

    # при поступлении числа в КА мы еще не понимаем к чему относится число,
    # поэтому сохраняем его в self._quantifier,
    # но как только мы его использовали,
    # то сразу его обнуляем для приему следующего числа
    _quantifier: int

    def __init__(self):
        # кол-во слов обработанных автоматом
        self._words_processed = 0
        self._is_period = False
        self._current_date = datetime.now()
        self._date_period = [None, None]
        self._relative_descriptor = 0
        self._quantifier = 0
        self._day = 0
        self._month = 0
        self._year = 0
        self._is_determined = False
        self._is_error = False
        self._next_expected_words = None

    def has_errors(self):
        return self._is_error

    @property
    def result(self) -> Tuple[Optional[str], Optional[str]]:
        """
        Возвращает результат работы конечного автомата:
        - кортеж из даты начала и окончания периода в случае успеха;
        - в случае ошибки одно из значений или оба могут быть ERROR_VALUE;
        - если дату не удалось распознать вернет пару UNRECOGNIZED_DATE_VALUE
        """

        if self._is_error:
            return ERROR_VALUE, ERROR_VALUE

        if self._is_determined:
            # если дата считается определенной,
            # но остался неиспользуемый квантификатор,
            # то это ошибка
            if self._quantifier:
                return ERROR_VALUE, ERROR_VALUE

            # если указали 1ую дату в периоде
            if self._date_period[0]:
                # если год определен, то надо его использовать
                if self._year:
                    try:
                        self._date_period[0] = safe_datetime(
                            year=self._year,
                            month=self._date_period[0].month,
                            day=self._date_period[0].day
                        )
                    except IncorrectDateException:
                        return ERROR_VALUE, ERROR_VALUE

                # если работаем с периодом или дата относительна текущей даты,
                # то конец периода всегда окончивается текущей датой
                # пример: "с 12 мая" или "за 4 прошлых месяца"
                if self._is_period or self._relative_descriptor:
                    # берем текущий день в качестве окончания
                    self._date_period[1] = self._current_date
                else:
                    # иначе берем расчитанную 2ую дату
                    if self._date_period[1]:
                        # если год определен, то надо его использовать
                        if self._year:
                            try:
                                self._date_period[1] = safe_datetime(
                                    year=self._year,
                                    month=self._date_period[1].month,
                                    day=self._date_period[1].day
                                )
                            except IncorrectDateException:
                                return ERROR_VALUE, ERROR_VALUE
                    else:
                        # если она пустая, то берем текущую дату
                        self._date_period[1] = self._current_date

                return format_date(self._date_period[0]),\
                       format_date(self._date_period[1])

            else:
                # Иначе строим дату начала на основе года, месяца и дня,

                # для частного случая когда указан только год
                if self._day == 0 and self._month == 0 and self._year:
                    self._date_period[0] = safe_datetime(self._year, 1, 1)

                    # если имеем дело с периодом когда
                    # в начале фразы союз "с" или "со"
                    if self._is_period:
                        # берем текущий день в качестве окончания
                        self._date_period[1] = self._current_date
                    else:
                        # иначе конец указанного года
                        self._date_period[1] = safe_datetime(self._year, 12, 31)

                    return format_date(self._date_period[0]),\
                           format_date(self._date_period[1])

                # если день не указан, тогда берем первый день
                if self._day == 0:
                    self._day = 1
                # если месяц не указан, тогда берем первый месяц
                if self._month == 0:
                    self._month = 1
                # если год не указан, тогда берем текущий год
                if self._year == 0:
                    self._year = self._current_date.year

                try:
                    self._date_period[0] = safe_datetime(self._year,
                                                         self._month,
                                                         self._day)
                except IncorrectDateException:
                    return ERROR_VALUE, ERROR_VALUE

                # берем текущий день в качестве окончания
                self._date_period[1] = self._current_date

                return format_date(self._date_period[0]), \
                       format_date(self._date_period[1])

        return UNRECOGNIZED_DATE_VALUE, UNRECOGNIZED_DATE_VALUE

    def input(self, word: str):
        """
        Ключевой метод КА принимающий слова
        и меняющий свое состояние в зависимости от этого
        """
        # считаем сколько слов обработал наш автомат
        self._words_processed += 1
        try:
            # если предлоги "от" или "с" идут не первыми словами,
            # то это ошибка
            if word == 'с' or word == 'со' or word == 'от':
                if self._words_processed == 1:
                    # зафиксировали что имеем дело с периодом
                    self._is_period = True
                    return
                else:
                    self._is_error = True
                    return

            # если в КА ошибка, то он больше не обрабатывает слова
            if self._is_error:
                return

            # если ожидаем какие-то слова
            if self._next_expected_words:
                # если встреченного слова нет в списке этих слов 
                if word not in self._next_expected_words:
                    self._is_error = True
                    return
                # иначе сбрасываем список в None и обрабатываем его
                self._next_expected_words = None
            # если _next_expected_words ИМЕННО пустой список,
            # то это означает, чтобольше не ждем никаких слов
            # и любое пришедшее слово является ошибкой
            elif self._next_expected_words == []:
                self._is_error = True
                return

            # проверка через регулярку дат формата dd.mm.yy
            m = re.match(re_short_date_pattern, word)
            if m:
                self._date_period[0] = self._date_period[1] = \
                    safe_datetime(
                        year=2000 + int(m.group(3)),
                        month=int(m.group(2)),
                        day=int(m.group(1))
                    )
                self._is_determined = True
                return

            # проверка через регулярку дат формата dd.mm.yyyy
            m = re.match(re_long_date_pattern, word)
            if m:
                self._date_period[0] = self._date_period[1] = \
                    safe_datetime(
                        year=int(m.group(3)),
                        month=int(m.group(2)),
                        day=int(m.group(1))
                    )
                self._is_determined = True
                return

            # проверка через регулярку дат формата dd.mm
            m = re.match(re_shortest_date_pattern, word)
            if m:
                self._date_period[0] = self._date_period[1] = \
                    safe_datetime(
                        year=self._current_date.year,
                        month=int(m.group(2)),
                        day=int(m.group(1))
                    )
                self._is_determined = True
                return

            # передаем число в КА
            if word.isnumeric():
                # квантификатор уже определен
                if self._quantifier:
                    self._is_error = True
                    return
                # иначе
                else:
                    self._quantifier = int(word)
                    if self._quantifier > 1900:
                        self._year = self._quantifier
                        self._quantifier = 0

            # иначе если в КА прилетело слово
            else:
                # год
                if match_word_with_list(word, ['год', 'лет']) != -1:
                    if self._quantifier:
                        # год должен быть 4х-значным и больще 1900 года
                        if self._quantifier > 1900:
                            self._year = self._quantifier
                            self._quantifier = 0
                    # указан относительный период
                    if self._relative_descriptor:
                        if self._relative_descriptor > 0:
                            self._year = self._current_date.year
                        else:
                            # квантификатор нужен для примера: 2 прошлых года
                            self._year = \
                                self._current_date.year - 1 \
                                * (self._quantifier if self._quantifier else 1)
                            self._quantifier = 0
                    else:
                        # если относительный период не указан,
                        # то период определяем как
                        # (365 * self._quanifier) дней ранее
                        # пример: за год - значит с периода 365 дней
                        # ранее по сегодня
                        if self._month == 0 and self._year == 0:
                            self._date_period[0] = \
                                self._current_date \
                                - timedelta(
                                    days=365 * (self._quantifier if self._quantifier else 1)
                                )
                            self._quantifier = 0

                    self._is_determined = True
                # месяц
                elif match_word_with_list(word, ['месяц']) != -1:
                    # указан относительный период
                    if self._relative_descriptor > 0:
                        self._month = self._current_date.month
                    elif self._relative_descriptor < 0:
                        # квантификатор нужен для примера: 2 прошлых месяца
                        delta_month = self._quantifier if self._quantifier else 0
                        if self._current_date.month < delta_month:
                            self._month = 12 + self._current_date.month - delta_month
                            self._year = self._current_date.year - 1
                        else:
                            self._month = self._current_date.month - delta_month
                            self._year = self._current_date.year
                        self._quantifier = 0
                    else:
                        # если относительный период не указан,
                        # то период определяем как (30 * self._quanifier)
                        # дней ранее
                        # пример:
                        # за месяц - значит с периода 30 дней ранее по сегодня
                        self._date_period[0] = \
                            self._current_date \
                            - timedelta(
                                days=30 * (self._quantifier if self._quantifier else 1)
                            )
                        self._quantifier = 0

                    self._is_determined = True
                # день
                elif match_word_with_list(word, ['ден', 'дня', 'дней']) != -1:
                    # указан относительный период
                    if self._relative_descriptor > 0:
                        self._date_period[0] = self._current_date
                    else:
                        # квантификатор нужен для примера: 2 прошлых дня
                        # когда квантификатор = 0, как в примере - за день,
                        # то это значит со вчерашнего дня
                        self._date_period[0] = self._current_date \
                            - timedelta(
                                days=self._quantifier if self._quantifier else 1
                            )

                        self._quantifier = 0

                    self._is_determined = True
                # неделя
                elif match_word_with_list(word, ['недел']) != -1:
                    # всегда относительный текущего дня период
                    self._date_period[0] = self._current_date \
                        - timedelta(
                            days=7 * (self._quantifier if self._quantifier else 1)
                        )

                    self._quantifier = 0

                    self._is_determined = True
                # квартал
                elif match_word_with_list(word, ['квартал']) != -1:
                    if self._quantifier:
                        if not (self._quantifier in [1, 2, 3, 4]):
                            self._is_error = True
                    else:
                        if self._current_date.month in [1, 2, 3]:
                            self._quantifier = 1
                        elif self._current_date.month in [4, 5, 6]:
                            self._quantifier = 2
                        elif self._current_date.month in [7, 8, 9]:
                            self._quantifier = 3
                        else:
                            self._quantifier = 4

                    self._date_period[0] = \
                        safe_datetime(
                            year=self._current_date.year,
                            month=3 * self._quantifier - 2,
                            day=1
                        )

                    if self._quantifier == 4:
                        self._date_period[1] = \
                            safe_datetime(
                                year=self._current_date.year + 1,
                                month=1,
                                day=1
                            ) - timedelta(days=1)
                    else:
                        self._date_period[1] = \
                            safe_datetime(
                                year=self._current_date.year,
                                month=3 * self._quantifier + 1,
                                day=1
                            ) - timedelta(days=1)

                    self._quantifier = 0
                    self._is_determined = True

                # другие слова
                else:
                    if match_word_with_list(word, ['вчерашн']) != -1:
                        self._next_expected_words = ['день']
                        self._relative_descriptor = -1
                    elif match_word_with_list(word, ['вчера']) != -1:
                        self._date_period[0] = self._date_period[1] = \
                            self._current_date - timedelta(days=1)
                        self._next_expected_words = []
                        self._is_determined = True
                    elif match_word_with_list(word, ['сегодн', 'ныне']) != -1:
                        self._date_period[0] = self._date_period[1] = \
                            self._current_date
                        self._next_expected_words = []
                        self._is_determined = True
                    elif match_word_with_list(word, ['сегодняшн']) != -1:
                        self._next_expected_words = ['день']
                        self._relative_descriptor = 1

                    else:
                        # обрабатваем месяца
                        months = [
                            'январ',
                            'феврал',
                            'март',
                            'апрел',
                            'ма',
                            'июн',
                            'июл',
                            'август',
                            'сентябр',
                            'октябр',
                            'ноябр',
                            'декабр'
                        ]
                        month_index = match_word_with_list(word, months)
                        if month_index != -1:
                            self._month = month_index + 1
                            if self._quantifier:
                                self._day = self._quantifier
                                self._quantifier = 0

                            if self._day:
                                self._date_period[0] = self._date_period[1] =\
                                    safe_datetime(
                                        year=self._current_date.year,
                                        month=self._month,
                                        day=self._day
                                    )
                            else:
                                self._date_period[0] = safe_datetime(
                                    year=self._current_date.year,
                                    month=self._month,
                                    day=1
                                )

                                self._date_period[1] = \
                                    safe_datetime(
                                        year=self._current_date.year + 1 if self._month == 12 else self._current_date.year,
                                        month=1 if self._month == 12 else self._month + 1,
                                        day=1
                                    ) - timedelta(days=1)

                            self._is_determined = True
                            return

                        pos_relative_descriptors = [
                            'текущ',
                            'нынешн',
                            'этот',
                            'эт'
                        ]
                        if match_word_with_list(word, pos_relative_descriptors) != -1:
                            self._relative_descriptor = 1
                            return

                        neg_relative_descriptors = [
                            'прошл',
                            'предыдущ'
                        ]
                        if match_word_with_list(word, neg_relative_descriptors) != -1:
                            self._relative_descriptor = -1
                            return

        except IncorrectDateException:
            self._is_error = True


def format_date(date: Optional[datetime]) -> str:
    """
    Форматирует переданое значение в строку,
    в случае если передали None возвращает ERROR_VALUE
    """
    return date.strftime(date_format) if date else ERROR_VALUE


def safe_datetime(year: int, month: int, day: int) -> Optional[datetime]:
    """
    Строит datetime из комбинации года, месяца и дня.
    Возбуждает исключение IncorrectDateException, в случае некорректной даты
    """
    try:
        return datetime(year=year, month=month, day=day)
    except ValueError as exc:
        raise IncorrectDateException('Некорректная дата') from exc


def match_word_with_list(word_to_check: str,
                         list_of_pattern_words: List[str]) -> int:
    """
    Проверяем слово на вхождение в список слов, указанных без окончания.
    Примеры:
    match_word_with_list("года", ["год"]) == True
    match_word_with_list("май", ["ма", "июн", "июл"]) == True
    match_word_with_list("мая", ["ма", "июн", "июл"]) == True
    match_word_with_list("зима", ["весн", "лет", "осен", "зим"]) == True

    :param word_to_check: проверяемое слово
    :param list_of_pattern_words: список слово без окончания
    :return: индекс слова совпадающего в списке, -1 - если совпадения нет
    """
    i: int = 0
    for pattern_word in list_of_pattern_words:
        # окончание это комбинация букв длинной до 3 символов
        # сначала думал что гласные только должны быть,
        # но окончание "ых" в слове "прошлых" заставило использова все буквы
        pattern_str: str = '^' + pattern_word + '[а-я]{0,3}$'
        if re.match(pattern_str, word_to_check):
            return i
        i += 1

    return -1


def is_from_date_dictionary(word: str) -> bool:
    """
    Функция определяет входит ли слово
    в "словарь" описывающий период времени
    """

    # некоторые слова лишены окончания для нивелирования влияния падежей
    list_of_dictionary: List[str] = [
        "от",
        "с",
        "со",
        "по",
        "до",
        "вчерашн",
        "вчера",
        "прошл",
        "текущ",
        "этот",
        "эт",
        "ныне",
        "нынешн",
        "сегодн",
        "сегодняшн",
        "день",
        "дня",
        "месяц",
        "год",
        "лет",
        "недел",
        "квартал",
        "январ",
        "феврал",
        "март",
        "апрел",
        "ма",
        "июн",
        "июл",
        "август",
        "сентябр",
        "октябр",
        "ноябр",
        "декабр"
    ]
    if match_word_with_list(word, list_of_dictionary) == -1:
        return False
    return True


def period_determiner(words: List[str],
                      max_days_in_period: Optional[int] = None,
                      future_days_allowed: bool = False) -> Tuple[str, str]:
    """
    Входная функция модуля, ее вызываем для получения дат.
    Она использует рабочую функцию date_determiner

    :param words: список слово в нижнем регистре
    :param max_days_in_period: максимальное количество дней в периоде
    :param future_days_allowed: нужно ли ограничивать период сегодняшним днем
    :return: кортеж дат в виде строки формата dd.mm.yyyy,
    если одна из них дата или обе ошибочные, то пара из ERROR_VALUE,
    а также пара из UNRECOGNIZED_DATE_VALUE - если дату не удалось распознать
    """

    # индекс слова "по" или "до" в списке переданных слов
    index_of_the_word_till = -1
    try:
        index_of_the_word_till = words.index('по')
    except ValueError as exc:
        pass

    if index_of_the_word_till == -1:
        try:
            index_of_the_word_till = words.index('до')
        except ValueError as exc:
            pass

    begin_of_period: str = UNRECOGNIZED_DATE_VALUE
    end_of_period: str = UNRECOGNIZED_DATE_VALUE

    # указаны начало и конец периода с помощью слов "по" или "до",
    if index_of_the_word_till != -1:
        begin_of_period, _ = date_determiner(words[:index_of_the_word_till])
        _, end_of_period = date_determiner(words[index_of_the_word_till + 1:])

        # в русском языке период можно выбрать еще вот как:
        # с 2 по 13 апреля 2021 - первый период номером дня,
        # а вторая дата в обучной форме.
        # Обрабатываем этот случай после КА
        if (begin_of_period == ERROR_VALUE or begin_of_period == '') and \
                end_of_period != '' and end_of_period != ERROR_VALUE:
            j: int = len(words[:index_of_the_word_till])
            if j <= 2:
                if words[j - 1].isnumeric():
                    d1 = int(words[j - 1])
                    m = re.match(re_long_date_pattern, end_of_period)
                    if m:
                        d2 = int(m.group(1))
                        m2 = m.group(2)
                        y2 = m.group(3)
                        if d1 > d2:
                            return ERROR_VALUE, ERROR_VALUE
                        else:
                            begin_of_period = '{}.{}.{}'\
                                .format(d1 if d1 > 9 else '0' + str(d1), m2, y2)
    # иначе имеем дело со словами, указывающих на единственную дату
    else:
        begin_of_period, end_of_period = date_determiner(words)

    # дата начала не должна быть больше даты окончания периода
    if re.match(re_long_date_pattern, begin_of_period) \
            and re.match(re_long_date_pattern, end_of_period):
        begin_date: datetime = datetime.strptime(begin_of_period, date_format)
        end_date: datetime = datetime.strptime(end_of_period, date_format)
        if begin_date > end_date:
            return ERROR_VALUE, ERROR_VALUE
        else:
            # контроль максимального количества дней в выбранном периоде
            if max_days_in_period:
                t: timedelta = end_date - begin_date
                if t.days > max_days_in_period:
                    return ERROR_VALUE, ERROR_VALUE
            # контроль будущих дат
            if not future_days_allowed:
                current_day: datetime = datetime.now()
                if begin_date > current_day:
                    return ERROR_VALUE, ERROR_VALUE
                if end_date > current_day:
                    end_of_period = format_date(current_day)

    return begin_of_period, end_of_period


def date_determiner(words: List[str]) -> Tuple[Optional[str], Optional[str]]:
    """
    Краеугольная функция всего модуля.
    Функция определяет дату на основе переданных слов на русском языке.
    Если в списке слов идет речь о конкретной дате,
    то функция вернет кортеж из этой даты   - date1, date1
    В случае если имеем дело с периодом, то - date1, date2

    :param words: список слов, описывающих дату или период
    :return: картеж дат в формате dd.mm.yyyy
    или ERROR_VALUE, ERROR_VALUE в случае ошибки
    """

    state_machine = StateMachineForDateDetermining()
    for word in words:
        state_machine.input(word)
        if state_machine.has_errors():
            return ERROR_VALUE, ERROR_VALUE

    return state_machine.result


def extract_words_describing_period(words_from_intent: List[str]) -> List[str]:
    """
    Функция извлекает список слов, описываюищих период,
    для последующей передачи функциям для определения периода

    :param words_from_intent: все слова из предложения в нижнем регистре
    :return: список слов описывающих период
    """

    words_to_process: List[str] = []
    for word in words_from_intent:
        if word.isnumeric() \
            or is_from_date_dictionary(word) \
                or re.match(re_shortest_date_pattern, word) \
                or re.match(re_short_date_pattern, word) \
                or re.match(re_long_date_pattern, word):
            words_to_process.append(word)

    return words_to_process




import unittest

from core.descriptions.descriptions_items import DescriptionsItems


class MockFactory():
    def __init__(self, id, items):
        self.id = id
        self.items = items
        self.custom_value = None

    def set_custom_value(self, value):
        self.custom_value = value


class DescriptionsTest(unittest.TestCase):
    def setUp(self):
        self.descr = DescriptionsItems(MockFactory, dict(id1=["raw_data_value1"], id2=["raw_data_value2"]))

    def test_getitem(self):
        obj1 = self.descr["id1"]
        assert obj1.items == ["raw_data_value1"]
        obj1.set_custom_value(1)
        obj2 = self.descr["id1"]
        assert obj2 == obj1

    def test_len(self):
        assert len(self.descr) == 2

    def test_iter(self):
        cnt = 0
        items = []
        for item in self.descr:
            cnt += 1
            items += self.descr[item].items
        assert cnt == 2 and set(items) == {"raw_data_value1", "raw_data_value2"}

    def test_get_keys(self):
        assert set(self.descr.get_keys()) == {"id1", "id2"}

    def test_has_key(self):
        assert "id1" in self.descr
        assert not "invalid" in self.descr

    def test_update_data(self):
        items = dict(id1=["raw_data_value11"], id3=["raw_data_value3"])
        expected = items
        self.descr.update_data(items)
        self.assertEqual(len(self.descr._items), 2)
        self.assertEqual(set(self.descr._raw_items), set(expected))

    def test_update_item(self):
        id = "id1"
        expected = "raw_data_value11"
        print(self.descr._raw_items)
        self.descr.update_item(id, expected)
        self.assertEqual(self.descr._raw_items[id], expected)

    def test_remove_item(self):
        id = "id1"
        self.descr.remove_item(id)
        self.assertIsNone(self.descr._items.get(id))
        self.assertIsNone(self.descr._raw_items.get(id))


from unittest import TestCase
from unittest.mock import Mock

from scenarios.scenario_models.field.field_filler_description import RegexpAndStringOperationsFieldFiller


class PickableMock(Mock):
    def __reduce__(self):
        return (Mock, ())


class PickablePicklableMock:
    pass


class TestRegexpStringOperationsFiller(TestCase):
    def setUp(self):
        self.items = {"exp": "1-[0-9A-Z]{7}"}

    def _test_operation(self, field_value, type_op, amount):
        self.items["operations"] = []
        text_preprocessing_result = PickablePicklableMock()
        text_preprocessing_result.original_text = field_value

        filler = RegexpAndStringOperationsFieldFiller(self.items)
        return filler._operation(field_value, type_op, amount)

    def test_operation(self):
        field_value = "1-RSAR09A"
        type_op = "lower"

        result = self._test_operation(field_value, type_op, None)
        self.assertEqual(field_value.lower(), result)

    def test_operation_amount(self):
        field_value = "1-RSAR09A"
        type_op = "lstrip"
        amount = "1-"

        result = self._test_operation(field_value, type_op, amount)
        self.assertEqual(field_value.lstrip(amount), result)

    def _test_extract(self, field_value):
        text_preprocessing_result = PickablePicklableMock()
        text_preprocessing_result.original_text = field_value

        filler = RegexpAndStringOperationsFieldFiller(self.items)
        return filler.extract(text_preprocessing_result, None)

    def test_extract_upper(self):
        field_value = "1-rsar09a"
        self.items["operations"] = [{"type":"upper"}]

        result = self._test_extract(field_value)
        self.assertEqual(field_value.upper(), result)

    def test_extract_rstrip(self):
        field_value = "1-RSAR09A !)"
        self.items["operations"] = [{"type":"rstrip", "amount": "!) "}]

        result = self._test_extract(field_value)
        self.assertEqual(field_value.rstrip("!) "), result)

    def test_extract_upper_rstrip(self):
        field_value = "1-rsar09a !)"
        self.items["operations"] = [ {"type":"upper"}, {"type":"rstrip", "amount": "!) "} ]

        result = self._test_extract(field_value)
        self.assertEqual(field_value.upper().rstrip("!) "), result)

    def test_extract_no_operations(self):
        field_value = "1-rsar09a !)"
        self.items["operations"] = []

        result = self._test_extract(field_value)
        self.assertIsNone(result)




from typing import Dict, Any, Optional

from smart_kit.action.http import HTTPRequestAction
from smart_kit.configs import settings


class SmartPayAction(HTTPRequestAction):
    url = None
    method = None

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        request_params = items.get("request_params")
        smartpay_params = items.get("smartpay_params")

        items = {
            "params": {"method": self.method, "url": self.url},
            "store": self.store,
            "behavior": items["behavior"]
        }
        if request_params:
            items["params"]["params"] = request_params
        if smartpay_params:
            items["params"]["json"] = smartpay_params

        super().__init__(items, id)


class SmartPayCreateAction(SmartPayAction):
    """Экшн для создания счёта

    Example::
        {
            "type": "smartpay_create",
            "behavior": "some_behavior", // обязательный параметр
            "smartpay_params": { // обязательноый параметр (см. https://developers.sber.ru/docs/ru/va/reference/smartservices/smartpay/processing/smartpay-api#создание-счета)
                "ptype": 1,
                "invoice": {
                    "purchaser": {
                        "email": "test@test.ru",
                        "phone": "71111111111",
                        "contact": "phone"
                    },
                    "delivery_info": {
                        "address": {
                            "country": "RU",
                            "city": "Москва",
                            "address": "Кутузовский проспект, 32"
                        },
                        "delivery_type": "courier",
                        "description": "Спросить '\''Где тут Сбердевайсы?'\''"
                    },
                    "invoice_params": [
                        {
                            "key": "packageName",
                            "value": "SbERdeVICEs"
                        },
                        {
                            "key": "packageKey",
                            "value": "ru.sberdevices.test"
                        }
                    ],
                    "order": {
                        "order_id": "a952b7ee-c928-4586-bd05-d932c21f749",
                        "order_number": "1952",
                        "order_date": "2021-04-07T08:24:37.729Z",
                        "service_id": "13",
                        "amount": 79900,
                        "currency": "RUB",
                        "purpose": "Покупка продуктов",
                        "description": "Заказ № 22-1952. Покупка продуктов",
                        "language": "ru-RU",
                        "tax_system": 0,
                        "order_bundle": [
                            {
                                "position_id": 1,
                                "name": "Транзистор",
                                "item_params": [
                                    {
                                        "key": "_itemAttributes_supplier_info.name",
                                        "value": "ООО Ромашка"
                                    },
                                    {
                                        "key": "_itemAttributes_supplier_info.inn",
                                        "value": "5009053292"
                                    },
                                    {
                                        "key": "_itemAttributes_nomenclature",
                                        "value": "Å\u001e13622200005881"
                                    },
                                    {
                                        "key": "_itemAttributes_agent_info.type",
                                        "value": "7"
                                    }

                                ],
                                "quantity": {
                                    "value": 1,
                                    "measure": "шт."
                                },
                                "item_amount": 79801,
                                "currency": "RUB",
                                "item_code": "21ff407a-fa98-11e8-80c5-0cc47a817926",
                                "item_price": 79801,
                                "discount_type": "amount",
                                "discount_value": 17687,
                                "tax_type": 6
                            },
                            {
                                "position_id": 2,
                                "name": "Тиристор",
                                "item_params": [
                                    {
                                        "key": "code",
                                        "value": "123значение321"
                                    },
                                    {
                                        "key": "pack",
                                        "value": "100"
                                    },
                                    {
                                        "key": "_itemAttributes_supplier_info.name",
                                        "value": "ООО Платежи"
                                    },
                                    {
                                        "key": "_itemAttributes_supplier_info.inn",
                                        "value": "7730253720"
                                    }
                                ],
                                "quantity": {
                                    "value": 1,
                                    "measure": "шт."
                                },
                                "item_amount": 99,
                                "currency": "RUB",
                                "item_code": "21ff407c-fa98-11e8-80c5-0cc47a817926",
                                "item_price": 99,
                                "discount_type": "amount",
                                "discount_value": 21,
                                "tax_type": 6,
                            }
                        ]
                    }
                }
            }
        }
    """
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        self.url = settings.Settings()["template_settings"]["smartpay_url"] + "/invoices"
        self.method = self.POST
        self.store = items.get("store", "smartpay_create_answer")
        super().__init__(items, id)


class SmartPayPerformAction(SmartPayAction):
    """Экшн для проведения платежа

    Example::
        {
            "type": "smartpay_perform",
            "behavior": "some_behavior", // обязательный параметр
            "invoice_id": "0", // обязательный параметр
            "smartpay_params": { // обязательный параметр (см. https://developers.sber.ru/docs/ru/va/reference/smartservices/smartpay/processing/smartpay-api#проведение-платежа)
                "user_id": {
                    "partner_client_id": "2223"
                },
                "operations": [
                    {
                        "operation": "payment",
                        "code": "invoice",
                        "value": "167900"
                    }
                ],
                "device_info": {
                    "device_platform_type": "iOS",
                    "device_platform_version": "13.6.1",
                    "device_model": "iPhone 7",
                    "device_manufacturer": "Apple",
                    "device_id": "83c3f257-46d8-41fe-951b-f79d04e288c2",
                    "surface": "SBOL",
                    "surface_version": "11.5.0"
                },
                "return_url": "https://ok",
                "fail_url": "https://fail"
            }
        }
    """
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        self.url = settings.Settings()["template_settings"]["smartpay_url"] + f"/invoices/{items['invoice_id']}"
        self.method = self.POST
        self.store = items.get("store", "smartpay_perform_answer")
        super().__init__(items, id)


class SmartPayGetStatusAction(SmartPayAction):
    """Экшн для получения статуса платежа

    см. https://developers.sber.ru/docs/ru/va/reference/smartservices/smartpay/processing/smartpay-api#получение-статуса
    Example::
        {
            "type": "smartpay_get_status",
            "behavior": "some_behavior", // обязательный параметр
            "invoice_id": "0", // обязательный параметр
            "inv_status": "executed", // опциональный параметр
            "wait": 50 // опциональный параметр
        }
    """
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        items["request_params"] = {}
        inv_status = items.get("inv_status")
        wait = items.get("wait")
        if inv_status:
            items["request_params"]["inv_status"] = inv_status
        if wait:
            items["request_params"]["wait"] = wait

        self.url = settings.Settings()["template_settings"]["smartpay_url"] + f"/invoices/{items['invoice_id']}"
        self.method = self.GET
        self.store = items.get("store", "smartpay_get_status_answer")
        super().__init__(items, id)


class SmartPayConfirmAction(SmartPayAction):
    """Экшн для подтверждения оплаты

    Example::
        {
            "type": "smartpay_confirm",
            "invoice_id": "0", // обязательный параметр
            "smartpay_params": { // опциональный параметр, задаётся для неполной суммы (см. https://developers.sber.ru/docs/ru/va/reference/smartservices/smartpay/processing/smartpay-api#подтверждение-оплаты)
                "invoice": {
                    "order": {
                        "amount": 79801,
                        "currency": "RUB",
                        "order_bundle": [
                            {
                                "position_id": 1,
                                "name": "Транзистор",
                                "quantity": {
                                    "value": 1,
                                    "measure": "шт."
                                },
                                "item_amount": 79801,
                                "currency": "RUB",
                                "item_code": "21ff407a-fa98-11e8-80c5-0cc47a817926",
                                "item_price": 79801
                            }
                        ]
                    }
                }
            }
        }
    """
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        self.url = settings.Settings()["template_settings"]["smartpay_url"] + f"/invoices/{items['invoice_id']}"
        self.method = self.PUT
        self.store = items.get("store", "smartpay_confirm_answer")
        super().__init__(items, id)


class SmartPayDeleteAction(SmartPayAction):
    """Экшн для отмены счёта

    см. https://developers.sber.ru/docs/ru/va/reference/smartservices/smartpay/processing/smartpay-api#отмена-счета
    Example::
        {
            "type": "smartpay_delete",
            "behavior": "my_behavior", // обязательный параметр
            "invoice_id": "0" // обязательный параметр
        }
    """
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        self.url = settings.Settings()["template_settings"]["smartpay_url"] + f"/invoices/{items['invoice_id']}"
        self.method = self.DELETE
        self.store = items.get("store", "smartpay_delete_answer")
        super().__init__(items, id)


class SmartPayRefundAction(SmartPayAction):
    """Экшн для возврата платежа

    Example::
        {
            "type": "smartpay_refund",
            "behavior": "some_behavior", // обязательный параметр
            "invoice_id": "0", // обязательный параметр
            "smartpay_params": { // опциональный параметр, задаётся при частичном возврате (см. https://developers.sber.ru/docs/ru/va/reference/smartservices/smartpay/processing/smartpay-api#частичный-возврат-платежа)
                "invoice": {
                    "order": {
                        "current_amount": 79900,
                        "refund_amount": 79801,
                        "currency": "RUB",
                        "order_bundle": [
                            {
                                "position_id": 1,
                                "name": "Транзистор",
                                "quantity": {
                                    "value": 1
                                },
                                "item_amount": 79801,
                                "item_code": "21ff407a-fa98-11e8-80c5-0cc47a817926"
                            }
                        ]
                    }
                }
            }
        }
    """
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        self.url = settings.Settings()["template_settings"]["smartpay_url"] + f"/invoices/{items['invoice_id']}"
        self.method = self.PATCH
        self.store = items.get("store", "smartpay_refund_answer")
        super().__init__(items, id)



# coding: utf-8
from typing import Union, Dict, Any, Optional, List

from core.basic_models.actions.basic_actions import Action
from core.basic_models.actions.command import Command
from core.model.base_user import BaseUser
from core.text_preprocessing.base import BaseTextPreprocessingResult


class CounterAction(Action):
    version: Optional[int]
    key: str
    value: int

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        items = items or {}
        self.key = items["key"]
        self.value = items.get("value", 1)
        self.lifetime = items.get("lifetime")


class CounterIncrementAction(CounterAction):
    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        user.counters[self.key].inc(self.value, self.lifetime)
        return commands


class CounterDecrementAction(CounterAction):
    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        user.counters[self.key].dec(-self.value, self.lifetime)
        return commands


class CounterClearAction(CounterAction):
    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        user.counters.clear(self.key)
        return commands


class CounterSetAction(CounterAction):
    version: Optional[int]
    key: str
    value: Optional[int]
    reset_time: bool
    time_shift: Optional[int]

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self.value = items.get("value")
        self.reset_time = items.get("reset_time", False)
        self.time_shift = items.get("time_shift", 0)

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        user.counters[self.key].set(self.value, self.reset_time, self.time_shift)
        return commands


class CounterCopyAction(Action):
    def __init__(self, items: Dict[str, Any], id: Optional[str] = None):
        super().__init__(items, id)
        self.src = items["source"]
        self.dst = items["destination"]
        self.reset_time = items.get("reset_time", False)
        self.time_shift = items.get("time_shift", 0)

    def run(self, user: BaseUser, text_preprocessing_result: BaseTextPreprocessingResult,
            params: Optional[Dict[str, Union[str, float, int]]] = None) -> List[Command]:
        commands = super().run(user, text_preprocessing_result, params)
        value = user.counters[self.src].value
        user.counters[self.dst].set(value, self.reset_time, self.time_shift)
        return commands



# coding: utf-8
from functools import cached_property

from core.basic_models.actions.basic_actions import Action
from core.model.factory import list_factory

from scenarios.scenario_models.field.field_descriptions.basic_field_description import BasicFieldDescription


class QuestionFieldDescription(BasicFieldDescription):
    DEFAULT_AVAILABLE_VALUE = True

    def __init__(self, items, id):
        super(QuestionFieldDescription, self).__init__(items, id)
        self._requests = items.get("questions", [])
        self._on_filled_actions = items.get("on_filled_actions", [])
        self._ask_again_requests = items.get("ask_again_questions", [])

    @cached_property
    @list_factory(Action)
    def ask_again_requests(self):
        return self._ask_again_requests



# coding: utf-8
import unittest
from unittest.mock import Mock

from smart_kit.handlers import handle_close_app
from smart_kit.utils.picklable_mock import PicklableMock


class TestScenarioDesc(dict):
    def form_type(self):
        return 'type'


class HandlerTest6(unittest.TestCase):
    def setUp(self):
        self.test_text_preprocessing_result = Mock('text_preprocessing_result')
        self.test_user = PicklableMock()
        self.test_user.name = "TestName"
        self.test_user.forms = Mock('forms')
        self.test_user.forms.remove_item = lambda x: 20
        self.test_user.descriptions = {"scenarios": {"any scenario": TestScenarioDesc()}}
        self.test_user.last_scenarios = Mock('last_scenarios')
        self.test_user.last_scenarios.last_scenario_name = "any scenario"
        self.test_user.last_scenarios.delete = lambda x: 10
        self.test_user.message.payload = {"skillId": 1, "intent": 2}
        self.test_user.message.message_name = "TestMessageName"
        self.test_user.message.device = PicklableMock()
        self.test_user.message.device.surface = "test_surface"
        self.test_user.message.channel = "test_channel"

        self.test_payload = {'message': {1: 1}}
        self.test_text_preprocessing_result.raw = 'any raw'
        self.app_name = "test_app"

    def test_handler_close_app_init(self):
        obj = handle_close_app.HandlerCloseApp(app_name=self.app_name)
        self.assertIsNotNone(obj.KAFKA_KEY)
        self.assertIsNotNone(obj._clear_current_scenario)

    def test_handler_close_app_run(self):
        self.assertIsNotNone(handle_close_app.log_const.KEY_NAME)
        obj = handle_close_app.HandlerCloseApp(app_name=self.app_name)
        self.assertEqual([], obj.run(self.test_payload, self.test_user))



import unittest
from unittest.mock import MagicMock

from core.request.rest_request import RestRequest
from smart_kit.utils.picklable_mock import PicklableMock


class RestRequestTest(unittest.TestCase):
    def setUp(self):
        self.expected = PicklableMock()

    def test_get(self):
        self.rr = RestRequest({"method": "get"})
        self.rr._requests_get = MagicMock(return_value=self.expected)
        data = ["text"]
        result = self.rr.run(data)
        self.rr._requests_get.assert_called_once_with(data)
        self.assertEqual(result, self.expected)

    def test_post(self):
        self.rr = RestRequest({})
        self.rr._requests_post = MagicMock(return_value=self.expected)
        data = ["text"]
        result = self.rr.run(data)
        self.rr._requests_post.assert_called_once_with(data)
        self.assertEqual(result, self.expected)

    def test_get_disabled(self):
        self.rr = RestRequest({"method": "get", "enabled": False})
        self.rr._requests_get = MagicMock(return_value=self.expected)
        data = ["text"]
        result = self.rr.run(data)
        self.assertIsNone(result)

    def test_post_disabled(self):
        self.rr = RestRequest({"enabled": False})
        self.rr._requests_post = MagicMock(return_value=self.expected)
        data = ["text"]
        result = self.rr.run(data)
        self.assertIsNone(result)



from time import time
from typing import List, Optional, Dict, Any

from core.basic_models.actions.command import Command
from core.logging.logger_utils import log

from core.text_preprocessing.preprocessing_result import TextPreprocessingResult
import scenarios.logging.logger_constants as log_const
from scenarios.user.user_model import User

from smart_kit.handlers.handler_base import HandlerBase
from smart_kit.names.action_params_names import TO_MESSAGE_NAME, SEND_TIMESTAMP
from core.monitoring.monitoring import monitoring


class HandlerRespond(HandlerBase):
    handler_name = "HandlerRespond"

    def __init__(self, app_name: str, action_name: Optional[str] = None):
        super().__init__(app_name)
        self._action_name = action_name

    def get_action_name(self, payload: Dict[str, Any], user: User):
        return self._action_name

    def get_action_params(self, payload: Dict[str, Any], user: User):
        callback_id = user.message.callback_id
        return user.behaviors.get_callback_action_params(callback_id)

    def run(self, payload: Dict[str, Any], user: User) -> List[Command]:
        commands = super().run(payload, user)
        callback_id = user.message.callback_id
        action_params = self.get_action_params(payload, user)
        action_name = self.get_action_name(payload, user)
        params = {
            log_const.KEY_NAME: "process_time",
            log_const.BEHAVIOR_CALLBACK_ID_VALUE: callback_id,
            "process_time": self.get_processing_time(user),
            "action_name": action_name,
            "surface": user.message.device.surface,
        }

        if user.behaviors.has_callback(callback_id):
            action_params = self.get_action_params(payload, user)
            params["to_message_name"] = action_params.get(TO_MESSAGE_NAME)
            log("HandlerRespond with action %(action_name)s started respond on %(to_message_name)s", user, params)
        else:
            log("HandlerRespond with action %(action_name)s started without callback", user, params)

        monitoring.counter_incoming(self.app_name, user.message.message_name, self.__class__.__name__, user)

        text_preprocessing_result = TextPreprocessingResult.from_payload(payload)

        if payload.get("message"):
            params = {
                log_const.KEY_NAME: log_const.NORMALIZED_TEXT_VALUE,
                "normalized_text": str(text_preprocessing_result.raw),
            }
            log("text preprocessing result: '%(normalized_text)s'", user, params, level="DEBUG")

        action = user.descriptions["external_actions"][action_name]
        commands.extend(action.run(user, text_preprocessing_result, action_params))
        return commands

    @staticmethod
    def get_processing_time(user: User):
        callback_id = user.message.callback_id
        process_time = None
        callback_params = user.behaviors.get_callback_action_params(callback_id) or {}
        if SEND_TIMESTAMP in callback_params:
            process_time = time() - callback_params[SEND_TIMESTAMP]
            process_time = int(process_time * 1000)
        return process_time



from unittest import TestCase
from unittest.mock import Mock

from scenarios.scenario_models.field.field import QuestionField
from smart_kit.utils.picklable_mock import PicklableMock, PicklableMagicMock


class TestField(TestCase):
    def test_1(self):
        expected = "my_value"
        mock_user = PicklableMock()
        mock_user.settings = {"template_settings": {}}
        description = PicklableMock()
        lifetime = 10
        items = {"value": expected, "available": False}
        field = QuestionField(description, items, mock_user, lifetime)

        result = field.value
        self.assertEqual(expected, result)

    def test_2(self):
        expected = "my_value"
        mock_user = PicklableMock()
        mock_user.settings = {"template_settings": {}}

        description = PicklableMock()
        description.need_load_context = False
        description.default_value = expected

        lifetime = 10
        items = None
        field = QuestionField(description, items, mock_user, lifetime)

        result = field.value
        self.assertEqual(expected, result)

    def test_3(self):
        expected = "prev_value"
        mock_user = PicklableMock()
        mock_user.last_fields = PicklableMagicMock()
        value_mock = PicklableMock()
        value_mock.value = "prev_value"
        mock_user.last_fields.__getitem__.return_value = value_mock
        mock_user.settings = {"template_settings": {}}

        description = PicklableMock()
        description.need_load_context = True
        description.id = 5

        lifetime = 10
        items = None
        field = QuestionField(description, items, mock_user, lifetime)

        result = field.value
        self.assertEqual(expected, result)

    def test_4(self):
        expected = "my_value"
        mock_user = PicklableMock()
        mock_user.last_fields = PicklableMagicMock()
        value_mock = PicklableMock()
        value_mock.value = None
        mock_user.last_fields.__getitem__.return_value = value_mock
        mock_user.settings = {"template_settings": {}}

        description = PicklableMock()
        description.need_load_context = True
        description.id = 5
        description.default_value = expected

        lifetime = 10
        items = None
        field = QuestionField(description, items, mock_user, lifetime)

        result = field.value
        self.assertEqual(expected, result)

    def test_5_1(self):
        mock_user = None
        description = PicklableMock()
        lifetime = 10
        items = {"value": "my_value", "available": True}
        field = QuestionField(description, items, mock_user, lifetime)

        self.assertTrue(field.valid)

    def test_5_2(self):
        mock_user = Mock(last_fields=[])

        description = Mock(default_value=None)
        description.id = 1
        mock_user.last_fields = {description.id: Mock(value=None)}
        mock_user.settings = {"template_settings": {}}

        description.required = False
        lifetime = 10
        items = {"available": True}
        field = QuestionField(description, items, mock_user, lifetime)

        self.assertTrue(field.valid)

    def test_5_3(self):
        mock_user = PicklableMock()

        description = Mock(default_value=None)
        description.id = 1
        mock_user.last_fields = {description.id: Mock(value=None)}
        mock_user.settings = {"template_settings": {}}
        lifetime = 10
        items = {"available": True}
        field = QuestionField(description, items, mock_user, lifetime)

        self.assertFalse(field.valid)

    def test_6_1(self):
        expected = {"value": "my_value", "available": True}

        mock_user = None
        description = PicklableMock()
        description.default_value = "some_value"
        description.available = False
        lifetime = 10
        items = {"value": "my_value", "available": True}
        field = QuestionField(description, items, mock_user, lifetime)

        result = field.raw
        self.assertDictEqual(expected, result)

    def test_6_2(self):
        expected = {"value": "my_value"}

        mock_user = None
        description = PicklableMock()
        description.default_value = "some_value"
        description.available = True
        lifetime = 10
        items = {"value": "my_value", "available": True}
        field = QuestionField(description, items, mock_user, lifetime)

        result = field.raw
        self.assertDictEqual(expected, result)

    def test_6_3(self):
        expected = {}
        mock_user = None
        description = PicklableMock()
        description.default_value = "some_value"
        description.available = True
        lifetime = 10
        items = {"value": None, "available": True}
        field = QuestionField(description, items, mock_user, lifetime)

        result = field.raw
        self.assertEqual(result, expected)

    def test_7_1(self):
        expected = "my_value"
        mock_user = None
        description = PicklableMock()
        description.default_value = "some_value"
        description.available = False
        lifetime = 10
        items = {"value": expected, "available": False}
        field = QuestionField(description, items, mock_user, lifetime)
        field.fill("some_value")
        result = field.value
        self.assertEqual(expected, result)
        self.assertFalse(field.available)

    def test_7_2(self):
        expected = "some_value"
        mock_user = None
        description = PicklableMock()
        description.default_value = "some_value"
        description.available = True
        description.need_save_context = False

        lifetime = 10
        items = {"value": "my_value", "available": True}
        field = QuestionField(description, items, mock_user, lifetime)
        field.fill("some_value")
        result = field.value
        self.assertEqual(expected, result)
        self.assertTrue(field.available)

    def test_7_3(self):
        expected = "some_value"

        mock_value = PicklableMock()
        mock_value.value = None

        mock_user = PicklableMagicMock()
        mock_user.last_fields = PicklableMagicMock()
        mock_user.last_fields.__getitem__.return_value = mock_value

        description = PicklableMock()
        description.id = 5
        description.default_value = expected
        description.available = True
        description.need_save_context = True

        lifetime = 10
        items = {"value": "my_value", "available": True}
        field = QuestionField(description, items, mock_user, lifetime)
        field.fill(expected)
        result = field.value

        self.assertEqual(expected, result)
        self.assertEqual(expected, mock_value.value)

    def test_7_4(self):
        expected = "my_value"
        mock_user = None
        description = PicklableMock()
        description.default_value = "some_value"
        description.available = True
        description.need_save_context = False
        description.need_load_context = False

        lifetime = 10
        items = {"value": "my_value", "available": True}
        field = QuestionField(description, items, mock_user, lifetime)
        field.fill(None)
        result = field.value
        self.assertEqual(expected, result)
        self.assertTrue(field.available)

    def test_set_available(self):
        mock_user = None
        description = PicklableMock()
        description.default_value = "some_value"
        description.available = True
        lifetime = 10
        items = {"value": None, "available": False}
        field = QuestionField(description, items, mock_user, lifetime)
        field.set_available()
        self.assertTrue(field.available)

    def test_set_unavailable(self):
        mock_user = None
        description = PicklableMock()
        description.default_value = "some_value"
        description.available = False
        lifetime = 10
        items = {"value": None, "available": True}
        field = QuestionField(description, items, mock_user, lifetime)
        field.reset_available()
        self.assertFalse(field.available)



import typing

from core.basic_models.actions.command import Command
from core.names import field

from scenarios.user.user_model import User

from smart_kit.names.message_names import ANSWER_TO_USER


def combine_answer_to_user(commands: typing.List[Command]) -> Command:
    from smart_kit.configs import get_app_config
    config = get_app_config()

    answer = Command(name=ANSWER_TO_USER, request_data=commands[0].request_data, request_type=commands[0].request_type)
    summary_pronounce_text = []
    auto_listening = None
    for command in commands:
        if command.request_data != answer.request_data:
            raise ValueError(f"Cant combine {ANSWER_TO_USER} commands, request_data is different")
        if command.request_type != answer.request_type:
            raise ValueError(f"Cant combine {ANSWER_TO_USER} commands, request_type is different")

        payload = command.payload

        pronounce_text = payload.pop(field.PRONOUNCE_TEXT, None)
        items = payload.pop(field.ITEMS, None)

        if auto_listening is None:
            auto_listening = payload.get(field.AUTO_LISTENING, None)

        if pronounce_text:
            summary_pronounce_text.append(pronounce_text)

        if items is not None:
            answer.payload.setdefault(field.ITEMS, []).extend(items)

        answer.payload.update(payload)

    if summary_pronounce_text:
        answer.payload[field.PRONOUNCE_TEXT] = " ".join(summary_pronounce_text)
    answer.payload[field.AUTO_LISTENING] = auto_listening

    return answer


def combine_commands(commands: typing.List[Command], user: User, **kwargs) -> typing.List[Command]:
    user_answers = []

    for command in commands:
        if command.name == ANSWER_TO_USER:
            user_answers.append(command)

    for user_answer in user_answers:
        commands.remove(user_answer)

    if user_answers:
        answer_to_user = combine_answer_to_user(user_answers)
        last_scenario_name = user.last_scenarios.last_scenario_name
        if field.INTENT not in answer_to_user.payload:
            if last_scenario_name is not None:
                answer_to_user.payload[field.INTENT] = last_scenario_name

        debug_info = answer_to_user.payload.setdefault(field.DEBUG_INFO, {})
        debug_info[field.INTENT] = last_scenario_name
        debug_info[field.DEBUG_INFO_APP_KEY] = user.message.app_info.project_id

        if field.FINISHED not in answer_to_user.payload:
            finished = last_scenario_name is None
            answer_to_user.payload[field.FINISHED] = finished

        from smart_kit.configs import get_app_config
        config = get_app_config()
        if answer_to_user.payload.get(field.AUTO_LISTENING) is None:
            answer_to_user.payload[field.AUTO_LISTENING] = config.AUTO_LISTENING

        events = user.history.get_events()
        if events and user.history.enabled:
            history_data = dict()
            history_data[field.EVENTS] = events
            answer_to_user.payload[field.HISTORY_DATA] = history_data

        commands.append(answer_to_user)  # Order isnt important
    return commands



# coding: utf-8
import logging

from core.logging.logger_utils import log


def run(app_config):
    log("RUN APP starting", level="WARNING")

    log("START SETTINGS CREATE", level="WARNING")
    settings = app_config.SETTINGS(
        config_path=app_config.CONFIGS_PATH, secret_path=app_config.SECRET_PATH,
        references_path=app_config.REFERENCES_PATH, app_name=app_config.APP_NAME)
    log("FINISHED SETTINGS CREATE", level="WARNING")
    source = settings.get_source()
    log("START RESOURCES CREATE", level="WARNING")
    resource = app_config.RESOURCES(source, app_config.REFERENCES_PATH, settings)
    log("FINISHED RESOURCES CREATE", level="WARNING")
    log("START MODEL CREATE", level="WARNING")
    model = app_config.MODEL(resource, app_config.DIALOGUE_MANAGER, settings)
    log("FINISHED MODEL CREATE", level="WARNING")

    log("START MAIN_LOOP CREATE", level="WARNING")
    loop = app_config.MAIN_LOOP(
        model, app_config.USER, app_config.PARAMETRIZER, app_config.POSTPROCESSOR_MAIN_LOOP,
        settings, app_config.TO_MSG_VALIDATORS, app_config.FROM_MSG_VALIDATORS,
    )
    log("FINISHED MAIN_LOOP CREATE", level="WARNING")
    loop.run()
    logging.shutdown()



# coding: utf-8
import unittest
from unittest.mock import Mock

from smart_kit.handlers import handler_timeout
from smart_kit.utils.picklable_mock import PicklableMock, PicklableMagicMock


class HandlerTest2(unittest.TestCase):
    def setUp(self):
        self.app_name = "TastAppName"
        self.test_user = Mock('user')
        self.test_user.id = '123-345-678'
        self.test_user.message = Mock('message')
        self.test_user.message.callback_id = 11
        self.test_user.message.incremental_id = 22
        self.test_user.message.logging_uuid = '321-654-987'
        self.test_user.message.channel = "channel"
        self.test_user.message.message_name = "test"
        self.test_user.message.app_info = None
        self.test_user.message.device = PicklableMock()
        self.test_user.message.device.surface = "surface"

        self.test_user.behaviors = Mock('behaviors')
        self.test_user.behaviors.timeout = lambda x: [120]
        self.test_user.behaviors.has_callback = lambda *x, **y: PicklableMagicMock()
        self.test_user.behaviors.get_callback_action_params = lambda *x, **y: {}
        self.test_payload = Mock('payload')

    def test_handler_timeout(self):
        obj = handler_timeout.HandlerTimeout(self.app_name)
        self.assertIsNotNone(obj.KAFKA_KEY)
        self.assertIsNotNone(handler_timeout.log_const.KEY_NAME)
        self.assertTrue(obj.run(self.test_payload, self.test_user) == [120])



import os
import objgraph
import psutil
import time


def get_meminfo():
    p = psutil.Process(os.getpid())
    return str(p.memory_info())


def show_growth(file=None):
    objgraph.show_growth(file=file, shortnames=False)


def show_most_common_types(file=None, limit=20):
    objgraph.show_most_common_types(limit=limit, shortnames=False, file=file)


def get_leaking_objects(file=None, limit=5):
    roots = objgraph.get_leaking_objects()
    objgraph.show_refs(roots[:limit], refcounts=True, shortnames=False, output=file)


if __name__ == "__main__":
    while 1 :
        print(show_most_common_types())
        time.sleep(1)



# coding: utf-8
import re

from core.logging.logger_constants import KEY_NAME
from core.logging.logger_utils import log

from prometheus_client import Counter, Histogram, REGISTRY


def _filter_monitoring_msg(msg):
    msg = msg.replace("-", ":")
    return msg


class MetricDisabled(ValueError):
    pass


def silence_it(func):
    def wrap(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except MetricDisabled as error:
            log(f"Metrics: {error}",
                params={KEY_NAME: "metrics_disabled"}, level="DEBUG")
        except:
            log("Metrics: Failed send. Exception occurred.",
                params={KEY_NAME: "metrics_fail"}, level="ERROR", exc_info=True)
    return wrap


class Monitoring:
    COUNTER = "counter"
    HISTOGRAM = "histogram"
    DEFAULT_ENABLED = True
    DEFAULT_DISABLED_METRICS = []

    def __init__(self):
        self._enabled = self.DEFAULT_ENABLED
        self.disabled_metrics = self.DEFAULT_DISABLED_METRICS.copy()
        self.buckets = Histogram.DEFAULT_BUCKETS
        self._monitoring_items = {
            self.COUNTER: {},
            self.HISTOGRAM: {}
        }
        self._clean_registry()

    @staticmethod
    def _clean_registry():
        """При создании нового инстанса мониторинга удаляем все созданные коллекторы"""
        collectors = list(REGISTRY._collector_to_names.keys())
        for collector in collectors:
            REGISTRY.unregister(collector)

    def check_enabled(self, name: str):
        metric_disabled = next((True for m in self.disabled_metrics if re.fullmatch(m, name)), False)
        return self._enabled and not metric_disabled

    def turn_on(self):
        self._enabled = True

    def turn_off(self):
        self._enabled = False

    def get_counter(self, name, description=None, labels=()):
        if not self.check_enabled(name):
            return None
        counter = self._monitoring_items[self.COUNTER]
        if not counter.get(name):
            counter[name] = Counter(name, description or name, labels)
        return counter[name]

    def got_counter(self, name, description=None, labels=()):
        counter = self.get_counter(name, description, labels)
        if counter:
            counter.inc()

    def got_histogram(self, name, description=None):
        def decor(func):
            def wrap(*args, **kwargs):
                decor_ = self._got_histogram(name, description=None)
                if decor_:
                    return decor_(func)(*args, **kwargs)
                else:
                    return func(*args, **kwargs)
            return wrap
        return decor

    def _got_histogram(self, name, description=None):
        if self.check_enabled(name):
            histogram = self._monitoring_items[self.HISTOGRAM]
            if not histogram.get(name):
                histogram[name] = Histogram(name, description or name, buckets=self.buckets)
            return histogram[name].time()

    def got_histogram_observe(self, name, value, description=None):
        if self.check_enabled(name):
            histogram = self._monitoring_items[self.HISTOGRAM]
            if not histogram.get(name):
                histogram[name] = Histogram(name, description or name, buckets=self.buckets)
            return histogram[name].observe(value)

    def apply_config(self, config):
        self._enabled = config.get("enabled", self.DEFAULT_ENABLED)
        self.disabled_metrics = config.get("disabled_metrics", self.DEFAULT_DISABLED_METRICS.copy())
        self.buckets = config.get("buckets", Histogram.DEFAULT_BUCKETS)

    @silence_it
    def init_metrics(self, app_name):
        self._get_or_create_counter(_filter_monitoring_msg("{}_load_error".format(app_name)), "Load user data error")
        self._get_or_create_counter(_filter_monitoring_msg("{}_save_error".format(app_name)), "Save user data error")
        self._get_or_create_counter(_filter_monitoring_msg("{}_save_collision".format(app_name)),
                                    "Save user data collision")
        self._get_or_create_counter(_filter_monitoring_msg("{}_save_collision_tries_left".format(app_name)),
                                    "Save user data collision all retries left.")
        self._get_or_create_counter(_filter_monitoring_msg("{}_exception".format(app_name)), "Exception in run-time.")
        self._get_or_create_counter(_filter_monitoring_msg("{}_invalid_message".format(app_name)),
                                    "Incoming message validation error.")

    def _get_or_create_counter(self, monitoring_msg, descr, labels=()):
        counter = monitoring.get_counter(monitoring_msg, descr, labels)
        if counter is None:
            raise MetricDisabled('counter disabled')
        return counter

    @silence_it
    def counter_incoming(self, app_name, message_name, handler, user, app_info=None):
        monitoring_msg = _filter_monitoring_msg("{}_incoming".format(app_name))

        c = self._get_or_create_counter(monitoring_msg, "Count of incoming messages",
                                        ['message_name', 'handler', 'project_id', 'system_name', 'application_id',
                                         'app_version_id', 'channel', 'surface'])
        if app_info is not None:
            project_id = app_info.project_id
            system_name = app_info.system_name
            application_id = app_info.application_id
            app_version_id = app_info.app_version_id
        else:
            project_id = system_name = application_id = app_version_id = None
        c.labels(message_name, handler, project_id, system_name,
                 application_id, app_version_id, user.message.channel, user.message.device.surface).inc()

    @silence_it
    def counter_outgoing(self, app_name, message_name, outgoing_message, user):

        monitoring_msg = _filter_monitoring_msg("{}_outgoing".format(app_name))

        c = self._get_or_create_counter(monitoring_msg, "Count of outgoing requests from application.",
                                        ['message_name', 'project_id', 'system_name', 'application_id',
                                         'app_version_id', 'channel', 'surface'])
        app_info = user.message.app_info
        c.labels(message_name, app_info.project_id, app_info.system_name,
                 app_info.application_id,
                 app_info.app_version_id, user.message.channel, user.message.device.surface).inc()

    @silence_it
    def counter_scenario_change(self, app_name, scenario, user):
        monitoring_msg = "{}_scenario_change".format(app_name)

        c = self._get_or_create_counter(monitoring_msg, "Count of scenario change events",
                                        ['scenario', 'project_id', 'system_name', 'application_id',
                                         'app_version_id', 'channel', 'surface'])
        app_info = user.message.app_info
        c.labels(scenario, app_info.project_id, app_info.system_name,
                 app_info.application_id,
                 app_info.app_version_id, user.message.channel, user.message.device.surface).inc()

    @silence_it
    def counter_nothing_found(self, app_name, scenario, user):
        monitoring_msg = "{}_outgoing_nothing_found".format(app_name)
        c = self._get_or_create_counter(monitoring_msg, "Count of scenario nothing found events",
                                        ['scenario', 'project_id', 'system_name', 'application_id',
                                         'app_version_id', 'channel', 'surface'])
        app_info = user.message.app_info
        c.labels(scenario, app_info.project_id, app_info.system_name,
                 app_info.application_id,
                 app_info.app_version_id, user.message.channel, user.message.device.surface).inc()

    @silence_it
    def counter_load_error(self, app_name):
        monitoring_msg = "{}_load_error".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg), "Load user data error")
        c.inc()

    @silence_it
    def counter_save_error(self, app_name):
        monitoring_msg = "{}_save_error".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg), "Save user data error")
        c.inc()

    @silence_it
    def counter_save_collision(self, app_name):
        monitoring_msg = "{}_save_collision".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg), "Save user data collision")
        c.inc()

    @silence_it
    def counter_save_collision_tries_left(self, app_name):
        monitoring_msg = "{}_save_collision_tries_left".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg),
                                        "Save user data collision all retries left.")
        c.inc()

    @silence_it
    def counter_exception(self, app_name):
        monitoring_msg = "{}_exception".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg), "Exception in run-time.")
        c.inc()

    @silence_it
    def counter_invalid_message(self, app_name):
        monitoring_msg = "{}_invalid_message".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg), "Incoming message validation error.")
        c.inc()

    @silence_it
    def counter_behavior_success(self, app_name, request_message_name):
        # 'Number of Success replies on messageName'
        self._behavior_monitoing_by_status(app_name, "success", request_message_name)

    @silence_it
    def counter_behavior_fail(self, app_name, request_message_name):
        # 'Number of Fail replies on messageName'
        self._behavior_monitoing_by_status(app_name, "fail", request_message_name)

    @silence_it
    def counter_behavior_misstate(self, app_name, request_message_name):
        # 'Number of Misstate replies on messageName'
        self._behavior_monitoing_by_status(app_name, "misstate", request_message_name)

    @silence_it
    def counter_behavior_timeout(self, app_name, request_message_name):
        # 'Number of Timeout replies on messageName'
        self._behavior_monitoing_by_status(app_name, "timeout", request_message_name)

    @silence_it
    def counter_behavior_expire(self, app_name, request_message_name):
        # 'Number of expire events on messageName'
        self._behavior_monitoing_by_status(app_name, "expire", request_message_name)

    def _behavior_monitoing_by_status(self, app_name, status, request_message_name):
        monitoring_msg = '{}_callback'.format(app_name)
        c = self._get_or_create_counter(monitoring_msg,
                                        "Count of incoming callback events with request_message_name",
                                        ['request_message_name', 'status'])

        c.labels(request_message_name, status).inc()

    @silence_it
    def counter_host_has_changed(self, app_name):
        monitoring_msg = '{}_host_has_changed'.format(app_name)
        c = self._get_or_create_counter(monitoring_msg,
                                        "Count of host has changed events within one message_id")
        c.inc()

    @silence_it
    def counter_mq_long_waiting(self, app_name):
        monitoring_msg = "{}_mq_long_waiting".format(app_name)
        c = self._get_or_create_counter(_filter_monitoring_msg(monitoring_msg),
                                        "(Now - creation_time) is greater than threshold")
        c.inc()

    @silence_it
    def sampling_load_time(self, app_name, value):
        monitoring_msg = "{}_load_time".format(app_name)
        monitoring.got_histogram_observe(_filter_monitoring_msg(monitoring_msg), value)

    @silence_it
    def sampling_script_time(self, app_name, value):
        monitoring_msg = "{}_script_time".format(app_name)
        monitoring.got_histogram_observe(_filter_monitoring_msg(monitoring_msg), value)

    @silence_it
    def sampling_save_time(self, app_name, value):
        monitoring_msg = "{}_save_time".format(app_name)
        monitoring.got_histogram_observe(_filter_monitoring_msg(monitoring_msg), value)

    @silence_it
    def sampling_mq_waiting_time(self, app_name, value):
        monitoring_msg = "{}_mq_waiting_time".format(app_name)
        monitoring.got_histogram_observe(_filter_monitoring_msg(monitoring_msg), value)


class MonitoringProxy:
    def __init__(self, default_cls):
        self.instance = default_cls()

    def got_histogram(self, *args_, **kwargs_):
        def decor_(func):
            def wrap(*args, **kwargs):
                wrapped_func = self.instance.got_histogram(*args_, **kwargs_)(func)
                value = wrapped_func(*args, **kwargs)
                return value
            return wrap
        return decor_

    def set_instance(self, cls):
        if type(self.instance) != cls:
            self.instance = cls()

    def __getattr__(self, item):
        return getattr(self.instance, item)


monitoring = MonitoringProxy(Monitoring)



from typing import Dict, Any

from core.basic_models.classifiers.basic_classifiers import classifier_factory
from core.descriptions.smart_updatable_descriptions_items import SmartUpdatableDescriptionsItems


class ExternalClassifiers(SmartUpdatableDescriptionsItems):

    def __init__(self, settings: Dict[str, Any]) -> None:
        super(ExternalClassifiers, self).__init__(classifier_factory, settings, ordered=True)



class BaseTextPreprocessingResult:

    def __init__(self, items):
        pass

    @property
    def tokenized_string(self):
        return NotImplementedError

    @property
    def tokenized_string_stop_words(self):
        return NotImplementedError

    @property
    def normalized_text_with_verb_mood(self):
        return NotImplementedError

    @property
    def words_tokenized(self):
        return NotImplementedError

    @property
    def words_tokenized_stop_words(self):
        return NotImplementedError

    @property
    def words_tokenized_set(self):
        return NotImplementedError

    @property
    def words_tokenized_set_stop_words(self):
        return NotImplementedError

    def get_token_values_by_type(self, token_type):
        return NotImplementedError

    def get_composite_token_values_by_type(self, composite_token_type):
        return NotImplementedError

    @property
    def num_token_values(self):
        return NotImplementedError

    @property
    def person_token_values(self):
        return NotImplementedError

    @property
    def money_token_values(self):
        return NotImplementedError

    @property
    def time_date_token_values(self):
        return NotImplementedError

    @property
    def period_token_values(self):
        return NotImplementedError

    @property
    def org_token_values(self):
        return NotImplementedError

    @property
    def geo_token_values(self):
        return NotImplementedError

    @property
    def time_date_interval_token_values(self):
        return NotImplementedError

    @property
    def time_time_token_values(self):
        return NotImplementedError

    @property
    def relative_token_values(self):
        return NotImplementedError

    @property
    def number_of_numbers(self):
        return NotImplementedError

    @property
    def ccy_token_values(self):
        return NotImplementedError

    @property
    def currencies_number(self):
        return NotImplementedError

    @property
    def tokenized_elements_list(self):
        return NotImplementedError

    @property
    def raw(self):
        return NotImplementedError



from unittest import TestCase

from core.model.registered import registered_factories
from scenarios.scenario_models.field.field_filler_description import FieldFillerDescription, CompositeFiller
from scenarios.scenario_models.field.field_filler_description import field_filler_factory, field_filler_description
from smart_kit.utils.picklable_mock import PicklableMock


class MockFiller:
    def __init__(self, items=None):
        items = items or {}
        self.result = items.get("result")

    def extract(self, text_preprocessing_result, user, params):
        return self.result


class TestCompositeFiller(TestCase):

    @classmethod
    def setUpClass(cls):
        registered_factories[FieldFillerDescription] = field_filler_factory
        field_filler_description["mock_filler"] = MockFiller
        TestCompositeFiller.user = PicklableMock()

    def test_first_filler(self):
        expected = "first"
        items = {
            "fillers": [
                {"type": "mock_filler", "result": expected},
                {"type": "mock_filler", "result": "second"}
            ]
        }
        filler = CompositeFiller(items)
        result = filler.extract(None, self.user)
        self.assertEqual(expected, result)

    def test_second_filler(self):
        expected = "second"
        items = {
            "fillers": [
                {"type": "mock_filler"},
                {"type": "mock_filler", "result": expected}
            ]
        }
        filler = CompositeFiller(items)
        result = filler.extract(None, self.user)
        self.assertEqual(expected, result)

    def test_not_fit(self):
        items = {
            "fillers": [
                {"type": "mock_filler"},
                {"type": "mock_filler"}
            ]
        }
        filler = CompositeFiller(items)
        result = filler.extract(None, self.user)
        self.assertIsNone(result)


from .formatters import EventFormatter, HistoryEventFormatter, formatters, formatters_factory
from .history import Event, History
from .history_description import HistoryDescription
from .constants import *



import argparse
import os
import shutil
import sys

import smart_kit
from core.descriptions.descriptions import Descriptions
from smart_kit.management.base import AppCommand
from smart_kit.testing.suite import TestSuite


def define_path(path):
    if os.path.exists(path):
        return path
    folder, _ = os.path.split(sys.argv[0])
    return os.path.join(folder, path)


class TestsCommand(AppCommand):
    TEST_TEMPLATE_PATH = "test_template.json"
    smart_kit_path = smart_kit.__path__[0]
    DEFAULT_TEMPLATE_PATH = os.path.join(smart_kit_path, "template/static/references/test_template.json")
    DEFAULT_PREDEFINED_FIELDS_STORAGE = os.path.join(
        smart_kit_path, "template/static/references/predefined_fields_storage.json")
    TEST_EXTENSION = ".json"

    def __init__(self, app_config):
        self.app_config = app_config
        self.parser = argparse.ArgumentParser(description="Tests creating and running.")
        self.parser.add_argument("path", metavar="PATH", type=str, help="Path to directory with tests", action="store")
        self.parser.add_argument("predefined_fields_storage", metavar="PREDEFINED_FIELDS_STORAGE", type=str,
                                 help="Path to json file with stored predefined fields", action="store", nargs="?",
                                 default=str(self.DEFAULT_PREDEFINED_FIELDS_STORAGE))
        self.commands = self.parser.add_mutually_exclusive_group(required=True)
        self.commands.add_argument("--run", dest="run", help="Runs Tests", action="store_true")
        run_command = self.parser.add_argument_group("Running")
        run_command.add_argument(
            "--make-csv", dest="make_csv", help="Create csv file for tests results", action="store_true"
        )

        self.commands.add_argument(
            "--gen", dest="gen", help="Create test directory at provided path", action="store_true"
        )
        gen_command = self.parser.add_argument_group("Generating")
        gen_command.add_argument(
            "--update", dest="update", help="Create missing template for scenarios", action="store_true",
        )

    def execute(self, *args, **kwargs):
        namespace = self.parser.parse_args(args)
        if namespace.gen:
            self.generate_tests_folder(namespace.path, namespace.update)
        elif namespace.run:
            self.run_tests(namespace.path, namespace.predefined_fields_storage, namespace.make_csv)
        else:
            raise Exception("Something going wrong due parsing the args")

    def generate_tests_folder(self, path: str, update=False):
        settings = self.app_config.SETTINGS(
            config_path=self.app_config.CONFIGS_PATH, secret_path=self.app_config.SECRET_PATH,
            references_path=self.app_config.REFERENCES_PATH, app_name=self.app_config.APP_NAME
        )
        resources = self.app_config.RESOURCES(settings.get_source(), self.app_config.REFERENCES_PATH, settings)
        scenario_names = Descriptions(resources.registered_repositories)["scenarios"].keys()
        folder_path = define_path(path)

        try:
            os.mkdir(folder_path)
            print(f"[+] Created tests folder at: {folder_path}")
        except FileExistsError:
            if not update:
                raise
            print(f"[+] Update tests folder at: {folder_path}")

        with open(self.get_test_template_path(), "r") as template_file:
            for scen_name in scenario_names:
                template_file.seek(0)
                new_file_path = os.path.join(folder_path, scen_name + self.TEST_EXTENSION)
                try:
                    with open(new_file_path, "x") as new_test_file:
                        shutil.copyfileobj(template_file, new_test_file)
                        print(f"[+] Created template file {new_file_path}")
                except FileExistsError:
                    if not update:
                        raise

    def run_tests(self, path, predefined_fields_storage, make_csv):
        path = define_path(path)
        predefined_fields_storage = define_path(predefined_fields_storage)
        if not os.path.exists(path):
            print(f"[!] Tests folder does not found at {path}")
            return
        elif not os.path.exists(predefined_fields_storage):
            print(f"[!] Predefined fields storage file does not found, check file path: {predefined_fields_storage}")
            return
        else:
            TestSuite(path, self.app_config, predefined_fields_storage, make_csv).run()

    def get_test_template_path(self):
        path = os.path.join(self.app_config.REFERENCES_PATH, self.TEST_TEMPLATE_PATH)
        if not os.path.exists(path):
            print(f"[!] Template for test file does not found. Expected at path {path}. Using default")
            path = self.DEFAULT_TEMPLATE_PATH
        return path



import re
import json
import time
import string
from typing import Dict

from core.text_preprocessing.constants import ANIMACY_TOKEN
from core.text_preprocessing.grammem.grammem_constants import TEXT, LEMMA, TOKEN_TYPE, LIST_OF_TOKEN_TYPES_DATA, \
    TOKEN_VALUE, VALUE, COMPOSITE_TOKEN_TYPE, IS_BEGINNING_OF_COMPOSITE, GRAMMEM_INFO, PART_OF_SPEECH, IS_STOP_WORD, \
    PART_OF_SPEECH_PYMORPHY

NUM_TOKEN = "NUM_TOKEN"
CUSTOM_PUNCTUATION = set(string.punctuation + "«»…=#-——–")
SENTENCE_ENDPOINT_TOKEN = "SENTENCE_ENDPOINT_TOKEN"


def flatten(lst: list) -> list:
    """
    Функция склеивает списки внутри списков, убирая один уровень вложенности
    Например, [ [1,2,3], [4,5,6], [[7,8],[9,10]] ] превращается в [1, 2, 3, 4, 5, 6, [7, 8], [9, 10]]
    """
    return [item for sublist in lst for item in sublist if item]


def reverse_json_dict(data):
    data = json.loads(data)
    result = dict()
    for key, values in data.items():
        for value in values:
            result[value] = key
    return result


def replace_by_dict(text: str, replace_rules: Dict[str, str]) -> str:
    for value, replacement in replace_rules.items():
        text = text.replace(value, replacement)
    return text


def revert_dict_with_list_values(dicti):
    new_dic = {}
    for k, v in dicti.items():
        for x in v:
            if x not in new_dic:
                new_dic[x] = k
    return new_dic


def replace_currencies_symbols(text: str) -> str:
    return replace_by_dict(text, {"$": " usd", "₽": " rur", "€": " eur"})


def extract_only_numeric(text: str) -> str:
    return ''.join([letter for letter in text if letter in "0123456789"])


def merge_numbers(text: str) -> str:
    """
    Класс предназначен для мержа номеров в одно число
    Например, "60 000 000" -> 60000000"

    :param text (string)
    :return: text(string)
    """
    tokens = text.split()
    result = tokens[0] if len(tokens) > 0 else ""
    for i in range(1, len(tokens)):
        if tokens[i].isdigit() and tokens[i - 1].isdigit():
            result += tokens[i]
        else:
            result = "{} {}".format(result, tokens[i])
    return result


def unmerge_numbers_and_letters(text: str) -> str:
    """
    Функция разбивает написанные вместе цифры и буквы. Например, "242р" -> "242 р"

    :param text (string)
    :return: result(string)
    """
    result = ''
    prev_letter = ''
    for letter in text:
        if (prev_letter.isalpha() and letter.isdigit()) or (prev_letter.isdigit() and letter.isalpha()):
            result = "{} {}".format(result, letter)
        else:
            result += letter
        prev_letter = letter
    return result


def lemma_modification(lemma: str) -> str:
    if len(lemma) > 1 and "." in lemma:
        res = lemma.strip(".")
        if "." in res:
            res = res.replace(".", "_")
    else:
        res = lemma
    return res


def return_lemmas_only(token_desc_list: list, include_sentence_endpoint: bool = True,
                       consider_stop_words: bool = False) -> str:
    """
    Конвертировать список токенов в строку без пунктуации, содержащую леммы и замены на тип токена.
    Предложения разделяются точкой. Если перед этим нет грамматической инфо - возвращается поле токена TEXT
    :param token_desc_list: Список токенов (дикты), include_sentence_endpoint -- принтить ли фейковую точку, bool
    :return: Строка, содержащую леммы и замены на тип токена.
    """
    final_line = []
    for token in token_desc_list:
        if COMPOSITE_TOKEN_TYPE in token:
            if token[IS_BEGINNING_OF_COMPOSITE]:
                token_type = token[COMPOSITE_TOKEN_TYPE]
            else:
                continue
        else:
            token_type = token.get(TOKEN_TYPE)
        if token_type != SENTENCE_ENDPOINT_TOKEN:
            grammem_info = token.get(GRAMMEM_INFO)
            if grammem_info is not None:
                pos = grammem_info.get(PART_OF_SPEECH_PYMORPHY)
                if pos != "PNCT":
                    if token_type is not None and token_type != ANIMACY_TOKEN:
                        final_line.append(token_type)
                    elif not token.get(IS_STOP_WORD) or (token.get(IS_STOP_WORD) and consider_stop_words):
                        lemma = lemma_modification(token.get(LEMMA))
                        final_line.append(lemma)
            else:
                final_line.append(token[TEXT])
        elif include_sentence_endpoint:
            final_line.append(token[TEXT])
    return " ".join(final_line)


class BaseNormalizePhoneNumbers(object):
    def __init__(self):
        self.boundary_start = '(?:[^\d\+]|^)'
        self.boundary_end = '(?:[^\d\-\:]|$)'
        self.plus_regex = re.compile(r"(\+ *)(\+7\d{10})")

    def remove_additional_phone_pluses(self, text: str) -> str:
        final_text = re.sub(self.plus_regex, r'\2', text)
        return final_text

    def _get_phone_numbers(self, text):
        return []

    def __call__(self, text: str) -> str:
        """
        На вход принимается строка
        На выходе имеем новую строку, где номера телефонов смержены в один токен

        :param text (string)
        :return: text(string)
        """
        already_done = []
        phone_numbers = self._get_phone_numbers(text)
        normalized_phone_numbers = []
        to_be_deleted = []
        for p in phone_numbers:
            digits = extract_only_numeric(p)[-10:]
            if digits[0] == "9":
                normalized_phone_numbers.append('+7' + digits)
            else:
                to_be_deleted.append(p)
        final_phone_numbers = [p for p in phone_numbers if p not in to_be_deleted]
        for normalized_pn, pn in zip(normalized_phone_numbers, final_phone_numbers):
            if (normalized_pn, pn.strip()) not in already_done:
                text = text.replace(pn.strip(), normalized_pn)
                already_done.append((normalized_pn, pn.strip()))
        text = self.remove_additional_phone_pluses(text)
        return text


class NormalizePhoneNumbers(BaseNormalizePhoneNumbers):
    def __init__(self):
        super(NormalizePhoneNumbers, self).__init__()
        self.phone_start = '(\+?[78]?'
        self.phone_delim = '[\s\-\(\)]?'
        self.phone_delim_with_dash = '[\) ][\- ]'
        self.phone_delim_dash_twice = '[\s\-\(\)]{,2}'

        t1 = "{}{}{}{}{}{}{}{}{}".format(self.phone_start, self.phone_delim, '\d{3}',
                                         self.phone_delim, '\d{3}', self.phone_delim, '\d{2}', self.phone_delim,
                                         '\d{2})')
        t2 = "{}{}{}{}{}{}{}{}{}".format(self.phone_start, self.phone_delim, '\d{4}',
                                         self.phone_delim, '\d{2}', self.phone_delim, '\d{2}', self.phone_delim,
                                         '\d{2})')
        t3 = "{}{}{}{}".format(self.phone_start, '\d{10}', self.phone_delim, ')')
        t4 = "{}{}{}{}{}{}{}{}{}".format(self.phone_start, self.phone_delim_dash_twice, '\d{3}',
                                         self.phone_delim_with_dash, '\d{3}', self.phone_delim, '\d{2}',
                                         self.phone_delim, '\d{2})')
        regex_phone_number_template = '|'.join([t1, t2, t3, t4])
        regex_phone_number_template = '{}(?:{}){}'.format(self.boundary_start, regex_phone_number_template,
                                                          self.boundary_end)

        self.regex_phone_number = re.compile(regex_phone_number_template)

    def _get_phone_numbers(self, text):
        return flatten(self.regex_phone_number.findall(text))


class NormalizePhoneNumbersVoice(BaseNormalizePhoneNumbers):
    def __init__(self):
        super(NormalizePhoneNumbersVoice, self).__init__()
        phone_start_1 = '\+?[78]'
        phone_start_2 = '9'
        phone_1 = '(([ \-]*\d){10})'
        phone_2 = '(([ \-]*\d){9})'

        t1 = "{}{}".format(phone_start_1, phone_1)

        t2 = "{}{}".format(phone_start_2, phone_2)

        regex_phone_number_template = '|'.join([t1, t2])

        regex_phone_number_template = '{}(?:{}){}'.format(self.boundary_start, regex_phone_number_template,
                                                          self.boundary_end)

        self.regex_phone_number = re.compile(regex_phone_number_template)

    def _get_phone_numbers(self, text):
        phone_numbers = []
        for z in self.regex_phone_number.finditer(text):
            phone_numbers.append(z.group(0))
        return phone_numbers


class MergeCardNumbers(object):
    def __init__(self):
        self._card_words = ["карта", "карты", "карте", "карту", "картой", "карт", "картах", "картам", "картами",
                            "виза", "visa", "мастеркард", "кард", "mastercard", "card", "maestro", "маэстро",
                            "мир"]
        self.words = "({}) ".format("|".join(self._card_words))
        self.card_regex = "((\d{4} \d{4} \d{4} \d{4})|(\d{4} \d{4} \d{10}))(?:(?!\d)|$)"
        self.regex_str = self.words + self.card_regex
        self._regex_card_number = re.compile(self.regex_str, re.IGNORECASE)

    def __call__(self, text: str) -> str:
        """
        Класс предназначен для мержа номеров карт в одно число при условии, что перед этим есть спец. слово
        Например, "карта 1234 1234 1234 1234" -> "карта 1234123412341234"

        :param text (string)
        :return: text(string)
        """
        for matched_tuple in self._regex_card_number.findall(text):
            card_word = matched_tuple[0]
            card_number = matched_tuple[1]
            text = text.replace("{} {}".format(card_word, card_number),
                                "{} {}".format(card_word, card_number.replace(' ', '')))
        return text


class MergeCardNumbersVoice(MergeCardNumbers):
    def __init__(self):
        super(MergeCardNumbersVoice, self).__init__()
        self.card_regex = '(([ ]*\d){16}(?:(?!\d)|$))'
        self.regex_str = self.words + self.card_regex
        self._regex_card_number = re.compile(self.regex_str, re.IGNORECASE)


class AdditionalMathSplitter(object):
    """
    Класс предназначен для дополнительного расспличивания по токенам для математического выражения.
    Деление уже ушло на этапе токенизации
    """

    def __init__(self):
        self.regexp = re.compile('(^|\s)(\d+([.,]\d+)?)(([*+\-^])(\d+([.,]\d+)?))+')
        self.str_of_delimiters = '*-^+'

    def __call__(self, text: str) -> str:
        searchen = re.search(self.regexp, text)
        if searchen:
            new_text = ''
            start_index = searchen.start()
            end_index = searchen.end()
            for i, char in enumerate(text):
                if start_index <= i <= end_index and char in self.str_of_delimiters:
                    new_text += " {} ".format(char)
                else:
                    new_text += char
        else:
            new_text = text
        return new_text


class UnicodeSymbolsConverter:
    """
    Замена юникодовых обозначений на человекочитаемые значки
    """

    def __init__(self, unicode_symbols: dict = None):
        self.unicode_dicti = unicode_symbols or dict()

    def __call__(self, text: str) -> str:
        '''
        :param text: str, текст, в котором будет произведена замена
        :return: текст, в котором юникодовая фигня заменена на человекочитаемые символы
        '''
        for key in self.unicode_dicti:
            text = text.replace(key, self.unicode_dicti[key])
        return text


class ReplaceSynonyms(object):
    """
    class to replace whole words in text avoiding replacements inside words based on regular expressions and synonyms dictionary from __init__
    example on "сбер": 'храните деньги в сберегательных кассах. сбер.' -> 'храните средства в сберегательных кассах. сбербанк.'
    requires lowered text without punctuation
    """

    def __init__(self, synonyms):
        self.synonyms = synonyms or {}

    def __call__(self, token_desc_list):
        """
        Класс предназначен для замены синонимов.
        На вход принимается список токенов
        На выходе имеем новый список токенов, где синонимы уже заменены

        :param token_desc_list (list of dicts)
        :return: token_desc_list (improved; list of dicts)
        """
        final_list_of_tokens = []
        skip = False
        for i, token in enumerate(token_desc_list):
            if skip:
                skip = False
            else:
                if len(token_desc_list) > i + 1 and \
                        "{} {}".format(token[TEXT].lower(), token_desc_list[i + 1][TEXT].lower()) in self.synonyms:
                    new_token = self.synonyms["{} {}".format(token[TEXT].lower(), token_desc_list[i + 1][TEXT].lower())]
                    if len(new_token.split(" ")) > 1:
                        to_extend = [{TEXT: token} for token in new_token.split(" ")]
                        final_list_of_tokens.extend(to_extend)
                    else:
                        final_list_of_tokens.append({TEXT: new_token})
                    skip = True
                elif token[TEXT].lower() in self.synonyms:
                    new_token = self.synonyms[token[TEXT].lower()]
                    if len(new_token.split(" ")) > 1:
                        to_extend = [{TEXT: token} for token in new_token.split(" ")]
                        final_list_of_tokens.extend(to_extend)
                    else:
                        final_list_of_tokens.append({TEXT: new_token})
                else:
                    final_list_of_tokens.append(token)
        return final_list_of_tokens


class CurrencyTokensOneIterationMerger(object):
    def __init__(self):
        self.currencies_list = ["rur", "коп", "копеек", "копейки", "копейка", "копейку", "копейками", "копейкой",
                                "копейкам",
                                "копейках", "копейке", "eur", "центом", "центу", "центе", "центах", "центами", "центы",
                                "usd", "aud", "cad", "цент", "цента", "центов", "sgd", "chf",
                                "jpy", "сена", "сены", "сен", "gbp", "пенни", "сантим", "сантима",
                                "сантимов", "dkk", "nok", "sek", "эре", "евроцент", "евроцента", "евроцентов",
                                "евроценту", "евроцентом", "евроценте", "евроцентах", "евроцентами", "евроценты",
                                "cny", "цзяо", "uah", "byn", "czk", "геллер", "геллера", "геллеров",
                                "egp", "пиастр", "пиастра", "пиастров", "try", "куруш", "куруша", "курушей",
                                "gel", "тетри", "amd", "лум", "сантим", "сантимах", "сантиму", "сантиме",
                                "сантимами", "сантима", "сантимов"]
        self.dicti_diminutives = {
            "usd": ["цент", "цента", "центов", "центом", "центу", "центе", "центах", "центами", "центы"],
            "eur": ["цент", "цента", "центов", "центом", "центу", "центе", "центах", "центами", "центы",
                    "евроцент", "евроцента", "евроцентов", "евроценту", "евроцентом", "евроценте", "евроцентах",
                    "евроцентами", "евроценты"],
            "rur": ["коп", "копеек", "копейки", "копейка", "копейку", "копейками", "копейках", "копейке", "копейкой",
                    "копейкам"],
            "sek": ["эре"],
            "dkk": ["эре"],
            "nok": ["эре"],
            "gbp": ["пенни"],
            "chf": ["сантим", "сантима", "сантимов", "сантимами", "сантимах", "сантиму", "сантиме", "сантимами"],
            "cad": ["цент", "цента", "центов", "центом", "центу", "центе", "центах", "центами", "центы"],
            "aud": ["цент", "цента", "центов", "центом", "центу", "центе", "центах", "центами", "центы"],
            "sgd": ["цент", "цента", "центов", "центом", "центу", "центе", "центах", "центами", "центы"],
            "jpy": ["сена", "сены"],
            "cny": ["цзяо"],
            "uah": ["коп", "копеек", "копейки", "копейка", "копейку", "копейками", "копейках", "копейке", "копейкой",
                    "копейкам"],
            "byn": ["коп", "копеек", "копейки", "копейка", "копейку", "копейками", "копейках", "копейке", "копейкой",
                    "копейкам"],
            "czk": ["геллер", "геллера", "геллеров"],
            "egp": ["пиастр", "пиастра", "пиастров"],
            "try": ["куруш", "куруша", "курушей"],
            "gel": ["тетри"],
            "amd": ["лум"]
        }
        self.final_small_currencies = flatten(list(self.dicti_diminutives.values()))
        self.denominator = 0.01
        self.backward_dict = revert_dict_with_list_values(self.dicti_diminutives)

    def _is_amount(self, tokenized_elements, i):
        return len(tokenized_elements) > i + 1 and tokenized_elements[i][TEXT].isdigit() and tokenized_elements[
            i + 1][TEXT].lower() in self.currencies_list

    def _denominator_are_ok(self, tokenized_elements, i):
        d1 = tokenized_elements[i + 1][TEXT].lower() if len(tokenized_elements) > i + 1 else None
        d2 = tokenized_elements[i - 1][TEXT].lower()
        if d2 in self.dicti_diminutives and d1 in self.dicti_diminutives[d2]:
            return True
        return False

    def _normalize_currency(self, tokenized_elements, i):
        if len(tokenized_elements) > i + 1 and tokenized_elements[i + 1][TEXT] in self.final_small_currencies:
            tokenized_elements[i][TEXT] = str(int(tokenized_elements[i][TEXT]) * self.denominator)
            tokenized_elements[i][TOKEN_VALUE][VALUE] *= self.denominator
            for dicti in tokenized_elements[i][LIST_OF_TOKEN_TYPES_DATA]:
                if dicti[TOKEN_TYPE] == NUM_TOKEN:
                    dicti[TOKEN_VALUE][VALUE] *= self.denominator
        return tokenized_elements[i]

    def _sum_2_amounts(self, final_tokenized_elements, tokenized_elements, i):
        final_tokenized_elements.append(self._normalize_currency(tokenized_elements, i))
        k = -1 if len(final_tokenized_elements) <= i else i
        final_tokenized_elements[k - 2][TEXT] = str(
            float(final_tokenized_elements[k - 2][TEXT]) + float(tokenized_elements[i][TEXT]))
        final_tokenized_elements[k - 2][TOKEN_VALUE][VALUE] += tokenized_elements[i][TOKEN_VALUE][VALUE]
        for dicti in final_tokenized_elements[k - 2][LIST_OF_TOKEN_TYPES_DATA]:
            if dicti[TOKEN_TYPE] == NUM_TOKEN:
                dicti[TOKEN_VALUE][VALUE] += tokenized_elements[i][TOKEN_VALUE][VALUE]
        final_tokenized_elements.pop(k)
        return final_tokenized_elements

    def __call__(self, tokenized_elements):
        """
        Класс предназначен для мержа валют -- переводить более мелкие (копейки и центы) в более крупные,
        а также мержить штуки вида 4 рубля 40 копеек в 4.4 рубля
        На вход принимается список токенов
        На выходе имеем список токенов, в котором смержены валюты

        :param tokenized_elements_list (list of dicts)
        :return: another tokenized_elements_list (list of dicts)
        """
        final_tokens = []
        skip, roubles = False, False
        for i in range(len(tokenized_elements)):
            if skip:
                skip = False
            elif roubles:
                other_currency = tokenized_elements[i][TEXT].lower()
                if other_currency in self.backward_dict:
                    final_tokens.append({TEXT: self.backward_dict[other_currency]})
                else:
                    final_tokens.append(tokenized_elements[i])
                roubles = False
            else:
                is_amount = self._is_amount(tokenized_elements, i)
                looks_like_2_amounts = is_amount and self._is_amount(tokenized_elements, i - 2) \
                                       and self._denominator_are_ok(tokenized_elements, i)
                if looks_like_2_amounts:
                    final_tokens = self._sum_2_amounts(final_tokens, tokenized_elements, i)
                    skip = True
                elif is_amount:
                    final_tokens.append(self._normalize_currency(tokenized_elements, i))
                    roubles = True
                else:
                    final_tokens.append(tokenized_elements[i])
        return final_tokens


class DateConverter:
    CENTURY = 100

    def __init__(self):
        self.regex = re.compile('(\d{1,2})[,\./\\\-](\d{1,2})[,\./\\\-](\d{4}|\d{2})')
        self.safe_regex_for_month_and_year = re.compile('(\d{1,2})[\\\/](\d{4}|\d{2})')

    def _year_preprocessing(self, year):
        tm_year_now = time.localtime().tm_year
        self.CENTURY_start = (tm_year_now // self.CENTURY) * self.CENTURY
        if year < self.CENTURY:
            if year <= 90:  # tm_year_now - self.CENTURY_start
                year += self.CENTURY_start
            else:
                year += self.CENTURY_start - self.CENTURY
        return year

    def __call__(self, word):
        r = self.regex.findall(word)
        safe_dates = self.safe_regex_for_month_and_year.findall(word)
        try:
            r = [int(i) for i in r[0]]
            r[2] = self._year_preprocessing(r[2])
            if r[1] > 12:
                r = [r[1], r[0], r[2]]
            result = {'day': r[0], 'month': r[1], 'year': r[2]}
        except (IndexError, ValueError):
            if safe_dates:
                r = [int(i) for i in safe_dates[0]]
                r[1] = self._year_preprocessing(r[1])
                result = {'month': r[0], 'year': r[1]}
            else:
                result = None

        return result


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]



from typing import Dict, Any, Optional

from core.model.base_user import BaseUser

from core.text_preprocessing.base import BaseTextPreprocessingResult

from core.basic_models.operators.operators import Operator
from core.basic_models.requirement.basic_requirements import ComparisonRequirement
from time import time


class CounterValueRequirement(ComparisonRequirement):
    operator: Operator
    key: str

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(CounterValueRequirement, self).__init__(items, id)
        items = items or {}
        self.key = items["key"]

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        counter = user.counters[self.key]
        return self.operator.compare(counter)


class CounterUpdateTimeRequirement(ComparisonRequirement):
    operator: Operator
    key: str

    def __init__(self, items: Dict[str, Any], id: Optional[str] = None) -> None:
        super(CounterUpdateTimeRequirement, self).__init__(items, id)
        items = items or {}
        self.key = items["key"]
        self.fallback_value = items.get("fallback_value") or False

    def check(self, text_preprocessing_result: BaseTextPreprocessingResult, user: BaseUser,
              params: Dict[str, Any] = None) -> bool:
        _time = user.counters[self.key].update_time
        return self.operator.compare(time() - _time) if _time else self.fallback_value



import json
import pickle
import tempfile
import unittest
from unittest.mock import patch, PropertyMock

from core.model.registered import Registered
from core.repositories.base_repository import BaseRepository
from core.repositories.classifier_repository import ClassifierRepository
from core.repositories.dill_repository import DillRepository
from core.repositories.folder_repository import FolderRepository
from core.repositories.shard_repository import ShardRepository
from core.utils.loader import ordered_json
from smart_kit.utils.picklable_mock import PicklableMock


class MockDescriptionItem:
    def __init__(self, value):
        self.value = value


class MockBaseRepository(BaseRepository):
    def __init__(self):
        super(MockBaseRepository, self).__init__(MockDescriptionItem)

    def load(self):
        self.fill({"test": {"value": 1}})
        super(MockBaseRepository, self).load()


class MockShardRepository(ShardRepository):
    def __init__(self):
        super(MockShardRepository, self).__init__(MockSource, ordered_json)


class MockSource:
    def __init__(self, content):
        self.content = content

    def list_dir(self, path):
        return self.content.keys()

    def open(self, *args, **kwargs):
        return MockStream(self.content[args[0]]['content'])


class MockStream:
    def __init__(self, data):
        self.data = data
        self.stream = PicklableMock()
        self.stream.read.return_value = json.dumps(self.data).encode()

    def __enter__(self):
        return self.stream

    def __exit__(self, *args):
        return None


class BaseRepositoryTest(unittest.TestCase):
    def test_base_repository_fill(self):
        data = {'test': {'value': 1}}
        repository = BaseRepository(key=None)
        repository.fill(data)
        self.assertEqual(repository.data, data)

    def test_base_repository_save(self):
        repository = BaseRepository(key=None)
        self.assertRaises(NotImplementedError, repository.save, {})

    def test_mock_repository(self):
        test_repository = MockBaseRepository()
        test_repository.load()
        self.assertEqual(test_repository.data, {'test': {'value': 1}})

    def test_base_repository_clear(self):
        test_repository = MockBaseRepository()
        test_repository.load()
        test_repository.clear()
        self.assertEqual(test_repository.data, {})


class ShardRepositoryTest(unittest.TestCase):
    def test_fill_on_top(self):
        shard_repository = MockShardRepository()
        shard_repository.fill({"test": {"value": 1}})
        self.assertEqual(shard_repository.data, {"value": 1})
        shard_repository.fill_on_top({"test": {"value2": 2}})
        self.assertEqual(shard_repository.data, {"value": 1, "value2": 2})


class RepositoryTest(unittest.TestCase):
    def test_repository(self):
        registered_repositories = Registered()
        obj = object()
        registered_repositories["test"] = obj
        self.assertEqual(registered_repositories["test"], obj)
        obj1 = object()
        registered_repositories["test"] = obj1
        self.assertEqual(registered_repositories["test"], obj1)


class FolderRepositoryTest(unittest.TestCase):
    def setUp(self):
        self.folder_content_test_dict = {'1.json': {'content': {'a': 'b'}}, '2.json': {'content': {'c': 'd'}}}
        self.folder_content_test_list = {'1.json': {'content': ['a', 'b']}, '2.json': {'content': ['c', 'd']}}
        self.folder_wrong_content = {'1.json': {'content': {'a': 'b'}}, '2.json': {'content': ['c', 'd']}}

    def test_right_filling_with_dict(self):
        test_repository = FolderRepository('', loader=json.loads,
                                           source=MockSource(self.folder_content_test_dict))
        test_repository.load()
        self.assertDictEqual(test_repository.data, {'a': 'b', 'c': 'd'})

    def test_right_filling_with_list(self):
        test_repository = FolderRepository('', loader=json.loads,
                                           source=MockSource(self.folder_content_test_list))
        test_repository.load()
        self.assertListEqual(test_repository.data, ['a', 'b', 'c', 'd'])

    def test_repo_wrong_content(self):
        test_repository = FolderRepository('', loader=json.loads,
                                           source=MockSource(self.folder_wrong_content))
        self.assertRaises(TypeError, test_repository.load)


class TestDillRepository(unittest.TestCase):
    def test_no_file_and_required(self):
        file = ''
        rep = DillRepository(filename=file)
        self.assertRaises(FileNotFoundError, rep.load)

    def test_no_file_and_not_required(self):
        file = ''
        rep = DillRepository(filename=file, required=False)
        rep.load()
        self.assertIsNone(rep.data)

    def test_file_exists(self):
        expected = {'a': 'b'}
        file = tempfile.NamedTemporaryFile(suffix='.pkl')
        with open(file.name, 'wb') as f:
            pickle.dump(expected, f)
        rep = DillRepository(filename=file.name)
        rep.load()
        self.assertEqual(expected, rep.data)


class TestClassifierRepository(unittest.TestCase):

    def setUp(self):
        self.temp_directory = tempfile.TemporaryDirectory()
        self.temp_directory_path = self.temp_directory.name

        self.temp_classifier_data = {"tests": "success"}
        self.temp_model_file = tempfile.NamedTemporaryFile(suffix=".pkl", dir=self.temp_directory_path)
        with open(self.temp_model_file.name, "wb") as f:
            pickle.dump(self.temp_classifier_data, f)
        self.temp_model_file_name = self.temp_model_file.name.split('/')[-1]

    def test_load_scikit_classifier(self):
        """Тест кейз на проверку загрузки моделей scikit классификаторов."""
        with patch(
                "core.repositories.folder_repository.FolderRepository.data",
                new_callable=PropertyMock,
                return_value={"test_classifier": {"type": "scikit", "path": self.temp_model_file_name, "intents": []}}
        ) as mock:
            with patch("core.repositories.folder_repository.FolderRepository.load") as load_mock:
                with patch("core.repositories.classifier_repository.classifiers_initial_launch") as initial_launch_mock:
                    classifier_repo = ClassifierRepository(
                        self.temp_directory_path, self.temp_directory_path, json.loads, "")
                    classifier_repo.load()
                    expected_result = {
                        "test_classifier": {
                            "classifier": {"tests": "success"},
                            "path": self.temp_model_file_name,
                            "type": "scikit",
                            "intents": []
                        }
                    }
                    self.assertEqual(expected_result, classifier_repo.data)

    def test_load_skip_classifier(self):
        """Тест кейз на проверку загрузки skip классификатора, для этого типа классификаторов сам
        файл модели не предусматривается."""
        expected_return_obj = {"test_classifier": {"type": "skip", "intents": []}}

        with patch("core.repositories.folder_repository.FolderRepository.data", new_callable=PropertyMock,
                   return_value=expected_return_obj) as mock_folder_data:
            with patch("core.repositories.folder_repository.FolderRepository.load") as load_mock:
                classifier_repo = ClassifierRepository(
                    self.temp_directory_path, self.temp_directory_path, json.loads, "")
                classifier_repo.load()
                self.assertEqual(expected_return_obj, classifier_repo.data)

    def test_load_external_classifier(self):
        """Тест кейз на проверку загрузки external классификатора."""
        expected_return_obj = {
            "test_classifier": {"type": "external", "classifier": "another_test_classifier"},
            "another_test_classifier": {"type": "scikit", "path": self.temp_model_file_name, "intents": []}
        }

        with patch("core.repositories.folder_repository.FolderRepository.data", new_callable=PropertyMock,
                   return_value=expected_return_obj) as mock_folder_data:
            with patch("core.repositories.folder_repository.FolderRepository.load") as load_mock:
                with patch("core.repositories.classifier_repository.classifiers_initial_launch") as initial_launch_mock:
                    classifier_repo = ClassifierRepository(
                        self.temp_directory_path, self.temp_directory_path, json.loads, "")
                    classifier_repo.load()
                    self.assertEqual(expected_return_obj, classifier_repo.data)

    def test_load_if_not_classifiers_paths(self):
        classifier_repo = ClassifierRepository("./nonexistent_path", "./another_nonexistent_path", json.loads, "")
        classifier_repo.load()
        self.assertIsNone(classifier_repo.data)

    def test_load_if_classifiers_directories_are_empty(self):
        new_empty_temp_directory = tempfile.TemporaryDirectory()
        empty_temp_directory_name = new_empty_temp_directory.name
        classifier_repo = ClassifierRepository(empty_temp_directory_name, empty_temp_directory_name, json.loads, "")
        classifier_repo.load()
        new_empty_temp_directory.cleanup()
        self.assertIsNone(classifier_repo.data)

    def tearDown(self):
        self.temp_directory.cleanup()



import time
import json


class ItemExpired(Exception):
    pass


class Cache:
    def __init__(self, lifetime: float = 0):
        self.lifetime = lifetime
        self.storage = dict()

    def __getitem__(self, item):
        create_time, value = self.storage[item]
        if create_time + self.lifetime <= time.time():
            raise ItemExpired
        return value

    def __setitem__(self, key, value):
        self.storage[key] = time.time(), value

    def load(self, *args, **kwargs):
        pass

    def save(self, *args, **kwargs):
        pass

    def invalidate(self):
        for key in list(self.storage):
            try:
                self[key]
            except ItemExpired:
                del self.storage[key]

    def clear(self):
        self.storage.clear()


class JSONCache(Cache):  # Pathetic Non OOP Design, Sorry
    def load(self, path):
        with open(path) as file:
            self.storage = json.load(file)

    def save(self, path, update=False):
        if update:
            tmp = JSONCache(lifetime=float("+inf"))
            tmp.load(path)
            tmp.storage.update(self.storage)
            self.storage = tmp.storage
        self.invalidate()
        with open(path, "w+", encoding='utf-8') as file:
            json.dump(self.storage, file, ensure_ascii=False)



# coding: utf-8
import unittest
from unittest.mock import Mock, MagicMock

from smart_kit.handlers import handler_take_profile_data


class MockVariables(Mock):
    _storage = {}

    def set(self, x, y):
        self._storage[x] = y

    def __getitem__(self, item):
        return self._storage[item]

    def get(self, item, default=None):
        if item in self._storage:
            return self._storage[item]
        return default


class HandleTakeProfileDataTest(unittest.TestCase):
    def setUp(self):
        self.app_name = "TestAppName"
        self.test_payload_1 = {"server_action": {}}
        self.test_payload_2 = {"server_action": {"action_id": 1, "parameters": 1}}
        self.test_user = MagicMock('user', message=MagicMock(message_name="some_name"), variables=MockVariables(),
                                   behaviors=MagicMock(success=lambda x: ["success"],
                                                       fail=lambda x: ["fail"],
                                                       timeout=lambda x: ["timeout"]))

    def test_handle_take_profile_data_init(self):
        obj = handler_take_profile_data.HandlerTakeProfileData(self.app_name)
        self.assertIsNotNone(obj.SUCCESS_CODE)
        self.assertIsNotNone(handler_take_profile_data.GEO)

    def test_handle_take_profile_data_run_fail(self):
        obj = handler_take_profile_data.HandlerTakeProfileData(self.app_name)
        payload = {"status_code": {"code": 102}}
        self.assertEqual(obj.run(payload, self.test_user), ["fail"])

    def test_handle_take_profile_data_run_success(self):
        obj = handler_take_profile_data.HandlerTakeProfileData(self.app_name)
        payload = {"profile_data": {"geo": {"reverseGeocoding": {"country": "Российская Федерация"},
                                            "location": {"lat": 10.125, "lon": 10.0124}}},
                   "status_code": {"code": 1}}
        self.assertEqual(obj.run(payload, self.test_user), ["success"])
        self.assertEqual(self.test_user.variables.get("smart_geo"), payload["profile_data"]["geo"])



# coding: utf-8
from typing import Dict, List, Optional, Any, Set

from core.basic_models.operators.operators import Operator
from core.model.factory import factory, build_factory, list_factory
from core.model.registered import Registered

field_requirements = Registered()

field_requirement_factory = build_factory(field_requirements)


class FieldRequirement:
    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        pass

    def check(self, field_value: str, params: Dict[str, Any] = None) -> bool:
        return True


class CompositeFieldRequirement(FieldRequirement):
    requirements: List[FieldRequirement]

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(CompositeFieldRequirement, self).__init__(items)
        self._requirements: List[Dict[str, Any]] = items["requirements"]
        self.requirements = self.build_requirements()

    @list_factory(FieldRequirement)
    def build_requirements(self):
        return self._requirements


class AndFieldRequirement(CompositeFieldRequirement):
    def check(self, field_value: str, params: Dict[str, Any] = None) -> bool:
        return all(requirement.check(field_value=field_value, params=params) for requirement in self.requirements)


class OrFieldRequirement(CompositeFieldRequirement):
    def check(self, field_value: str, params: Dict[str, Any] = None) -> bool:
        return any(requirement.check(field_value=field_value, params=params) for requirement in self.requirements)


class NotFieldRequirement(FieldRequirement):
    requirement: FieldRequirement

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(NotFieldRequirement, self).__init__(items)
        self._requirement: Dict[str, Any] = items["requirement"]
        self.requirement = self.build_requirement()

    @factory(FieldRequirement)
    def build_requirement(self):
        return self._requirement

    def check(self, field_value: str, params: Dict[str, Any] = None) -> bool:
        return not self.requirement.check(field_value=field_value, params=params)


class ComparisonFieldRequirement(FieldRequirement):
    operator: Operator

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(ComparisonFieldRequirement, self).__init__(items)
        self._operator: Dict[str, Any] = items["operator"]
        self.operator = self.build_operator()

    @factory(Operator)
    def build_operator(self):
        return self._operator

    def check(self, field_value: str, params: Dict[str, Any] = None) -> bool:
        return self.operator.compare(field_value)


class IsIntFieldRequirement(FieldRequirement):
    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(IsIntFieldRequirement, self).__init__(items)

    def check(self, field_value: str, params: Dict[str, Any] = None) -> bool:
        try:
            int(field_value)
            return True
        except ValueError:
            return False


class ValueInSetRequirement(FieldRequirement):
    symbols: List

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(ValueInSetRequirement, self).__init__(items)
        self.symbols: Set = set(items["symbols"])

    def check(self, field_value: str, params: Dict[str, Any] = None) -> bool:
        return field_value in self.symbols


class TokenPartInSet(FieldRequirement):
    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(TokenPartInSet, self).__init__(items)
        self.part = items['part']
        self.values = items['values']

    def check(self, field_value: dict, params: Dict[str, Any] = None) -> bool:
        return field_value[self.part] in self.values


class TextLengthFieldRequirement(FieldRequirement):
    min_field_length: int
    max_field_length: int

    def __init__(self, items: Optional[Dict[str, Any]]) -> None:
        super(TextLengthFieldRequirement, self).__init__(items)
        self.min_field_length = items["min_field_length"]
        self.max_field_length = items["max_field_length"]

    def check(self, field_value: str, params: Dict[str, Any] = None) -> bool:
        return self.min_field_length <= len(field_value) <= self.max_field_length



# coding: utf-8


class Field:
    def __init__(self, name, model, description=None, *args):
        self.name = name
        self.model = model
        self.description = description
        self.args = args



from time import time
from typing import Dict, Any, Optional, NamedTuple, List, Union, TYPE_CHECKING

from core.logging.logger_utils import log

if TYPE_CHECKING:
    from scenarios.scenario_models.history import HistoryDescription
    from scenarios.user.user_model import User


class Event(NamedTuple):
    type: Optional[str] = None
    scenario: Optional[str] = None
    scenario_version: Optional[str] = None
    node: Optional[str] = None
    result: Optional[str] = None
    content: Optional[Dict[str, str]] = None
    created_time: float = time()

    def to_dict(self) -> Dict[str, Any]:
        return self._asdict()


class History:

    _events: List[Event]
    _description: 'HistoryDescription'

    def __init__(self, items: Dict[str, Any], description: 'HistoryDescription', user: 'User'):
        items = items or {}
        self._description = description
        self._events = [Event(**e) for e in items.get('events', [])]
        if not self._description.enabled:
            log("History: scenario history events logging disabled", level="WARNING")

    @property
    def raw(self) -> Dict[str, Any]:
        return {
            "events": [e.to_dict() for e in self.get_raw_events()]
        }

    @property
    def enabled(self) -> bool:
        return self._description.enabled

    def get_events(self) -> List[Union[NamedTuple, Dict[str, Any]]]:
        return self._description.formatter.format(self._events)

    def get_raw_events(self) -> List[Event]:
        return self._events

    def add_event(self, event: Event):
        if self.enabled:
            self._events.append(event)

    def clear(self):
        self._events.clear()

    def expire(self):
        now = time()
        non_expired = []
        for event in self._events:
            if event.created_time + self._description.event_expiration_delay > now:
                non_expired.append(event)
        self._events = non_expired
